from __future__ import annotations  # Support new typing structure in 3.8 and 3.9

import copy
import logging
import math
import pickle
from dataclasses import dataclass, field
from math import floor
from pathlib import Path
from typing import Iterable, List, Optional, Union

import datasets
import numpy as np
from datasets import Features, IterableDataset
from escnn.group import Representation, groups_dict

log = logging.getLogger(__name__)


@dataclass
class DynamicsRecording:
    """Data structure to store recordings of a Markov Dynamics."""

    description: Optional[str] = None
    info: dict[str, object] = field(default_factory=dict)
    dynamics_parameters: dict = field(default_factory=lambda: {"dt": None})

    # Ordered list of observations composing to the state and action space of the Markov Process.
    state_obs: tuple[str, ...] = field(default_factory=list)
    action_obs: tuple[str, ...] = field(default_factory=list)

    # Map from observation name to the observation representation name. This name should be in `group_representations`.
    obs_representations: dict[str, Representation] = field(default_factory=dict)
    recordings: dict[str, Iterable] = field(default_factory=dict)
    # Map from observation name to the observation moments (mean, var) of the recordings.
    obs_moments: dict[str, tuple] = field(default_factory=dict)

    def __post_init__(self):
        # Check provided state observables are in the recordings
        for obs_name in self.state_obs:
            assert obs_name in self.recordings.keys(), (
                f"State observable {obs_name} not in the provided recordings: {self.recordings.keys()}"
            )
        for obs_name in self.recordings.keys():
            # Scalar observations should have shape (traj, time, 1)
            # obs_shape = self.recordings[obs_name].shape
            # if len(obs_shape) == 2:
            #     self.recordings[obs_name] = self.recordings[obs_name][..., None]
            # If group representation of observation is provided, check the representation has the same dimension
            if obs_name in self.obs_representations.keys():
                rep = self.obs_representations[obs_name]
                assert rep.size == self.recordings[obs_name].shape[-1], (
                    f"[{obs_name}] Invalid rep dim={rep.size} for {self.recordings[obs_name]}"
                )

    @property
    def obs_dims(self):
        """Dictionary providing the map between observation name and observation dimension."""
        return {k: v.shape[-1] for k, v in self.recordings.items()}

    @property
    def state_dim(self):
        """Return the dimension of the state vector."""
        return sum([self.obs_dims[obs] for obs in self.state_obs])

    _obs_idx = None

    def obs_idx_in_state(self, obs_name) -> tuple[int]:
        """Returns a tuple of indices of each observation in the state vector."""
        assert obs_name in self.state_obs, f"Observation {obs_name} not in state observations"
        if self._obs_idx is None:
            self._obs_idx = {}
            # Iterate over all state observables in order and store the indices of each observation given its dimension
            curr_dim = 0
            for obs in self.state_obs:
                obs_dim = self.obs_dims[obs]
                self._obs_idx[obs] = tuple(range(curr_dim, curr_dim + obs_dim))
                curr_dim += obs_dim
        return self._obs_idx[obs_name]

    def state_representations(self) -> list[Representation]:
        """Return the ordered list of representations of the state vector."""
        return [self.obs_representations[m] for m in self.state_obs]

    def state_moments(self) -> [np.ndarray, np.ndarray]:
        """Compute the mean and standard deviation of the state observations."""
        mean, var = [], []
        for obs_name in self.state_obs:
            if obs_name not in self.obs_moments.keys():
                self.compute_obs_moments(obs_name)
            obs_mean, obs_var = self.obs_moments[obs_name]
            mean.append(obs_mean)
            var.append(obs_var)
        mean, var = np.concatenate(mean), np.concatenate(var)
        return mean, var

    def compute_obs_moments(self, obs_name: str) -> [np.ndarray, np.ndarray]:
        """Compute the mean and standard deviation of observations."""
        assert obs_name in self.recordings.keys(), f"Observation {obs_name} not found in recording"
        is_symmetric_obs = obs_name in self.obs_representations.keys()
        if is_symmetric_obs:
            rep_obs = self.obs_representations[obs_name]
            obs_original_basis = np.asarray(self.recordings[obs_name])
            G = rep_obs.group
            # Allocate the mean and variance arrays.
            mean, var = np.zeros(rep_obs.size), np.ones(rep_obs.size)
            # Change basis of the observation to expose the irrep G-stable subspaces
            Q_inv = rep_obs.change_of_basis_inv  # Orthogonal transformation to irrep basis (Q^T = Q^-1)
            Q = rep_obs.change_of_basis
            # Get the dimensions of each irrep.

            S = np.zeros((rep_obs.size, rep_obs.size))
            irreps_dimension = []
            cum_dim = 0
            for irrep_id in rep_obs.irreps:
                irrep = G.irrep(*irrep_id)
                # Get dimensions of the irrep in the original basis
                irrep_dims = range(cum_dim, cum_dim + irrep.size)
                irreps_dimension.append(irrep_dims)
                if irrep_id == G.trivial_representation.id:
                    S[irrep_dims, irrep_dims] = 1
                cum_dim += irrep.size

            # Compute the mean of the observation.
            # The mean of a symmetric random variable (rv) lives in the subspaces associated with the trivial irreps.
            has_trivial_irreps = G.trivial_representation.id in rep_obs.irreps
            if has_trivial_irreps:
                avg_projector = rep_obs.change_of_basis @ S @ rep_obs.change_of_basis_inv
                # Compute the mean in a single vectorized operation
                mean_empirical = np.mean(obs_original_basis, axis=(0, 1))
                mean = np.einsum("...ij,...j->...i", avg_projector, mean_empirical)

            # Compute the variance of the observable by computing a single variance per irrep G-stable subspace.
            # To do this, we project the observations to the basis exposing the irreps, compute the variance per
            # G-stable subspace, and map the variance back to the original basis.
            centered_obs_irrep_basis = np.einsum("...ij,...j->...i", Q_inv, obs_original_basis - mean)
            var_irrep_basis = np.ones_like(var)
            for irrep_id, irrep_dims in zip(rep_obs.irreps, irreps_dimension):
                irrep = G.irrep(*irrep_id)
                centered_obs_irrep = centered_obs_irrep_basis[..., irrep_dims]
                assert centered_obs_irrep.shape[-1] == irrep.size, (
                    f"Obs irrep shape {centered_obs_irrep.shape} != {irrep.size}"
                )

                # Since the irreps are unitary/orthogonal transformations, we are constrained compute a unique variance
                # for all dimensions of the irrep G-stable subspace, as scaling the dimensions independently would break
                # the symmetry of the rv. As a centered rv the variance is the expectation of the squared rv.
                var_irrep = np.mean(centered_obs_irrep**2)  # Single scalar variance per G-stable subspace
                # Store the irrep mean and variance in the entire representation mean and variance
                var_irrep_basis[irrep_dims] = var_irrep
            # Convert the variance from the irrep basis to the original basis
            Cov = Q @ np.diag(var_irrep_basis) @ Q_inv
            var = np.diagonal(Cov)

            # TODO: Move this check to Unit test as it is computationally demanding to check this at runtime.
            # Ensure the mean is equivalent to computing the mean of the orbit of the recording under the group action
            # aug_obs = []
            # for g in G.elements:
            #     g_obs = np.einsum('...ij,...j->...i', rep_obs(g), obs_original_basis)
            #     aug_obs.append(g_obs)
            #
            # aug_obs = np.concatenate(aug_obs, axis=0)   # Append over the trajectory dimension
            # mean_emp = np.mean(aug_obs, axis=(0, 1))
            # assert np.allclose(mean, mean_emp, rtol=1e-3, atol=1e-3), f"Mean {mean} != {mean_emp}"
            #
            # var_emp = np.var(aug_obs, axis=(0, 1))
            # assert np.allclose(var, var_emp, rtol=1e-2, atol=1e-2), f"Var {var} != {var_emp}"
        else:
            mean = np.mean(np.asarray(self.recordings[obs_name]), axis=(0, 1))
            var = np.var(np.asarray(self.recordings[obs_name]), axis=(0, 1))
        assert mean.shape == (self.obs_dims[obs_name],), (
            f"Obs {obs_name} dim ({self.obs_dims[obs_name]},) diff from estimated mean dim ({mean.shape},)!= "
        )
        assert var.shape == (self.obs_dims[obs_name],), (
            f"Obs {obs_name} dim ({self.obs_dims[obs_name]},) diff from estimated var dim ({var.shape},)!= "
        )

        self.obs_moments[obs_name] = mean, var

    def get_state_trajs(self, standardize: bool = False):
        """Returns a single array containing the concatenated state observations trajectories.

        Given the state observations `self.state_obs` this method concatenates the trajectories of each observation
        into a single array of shape [traj, time, state_dim]. If standardize is set to True, the state observations
        are standardized to have zero mean and unit variance.

        Returns:
            A single array containing the concatenated state observations trajectories of shape [traj, time, state_dim].
        """
        obs = [self.recordings[obs_name] for obs_name in self.state_obs]
        state_trajs = np.concatenate(obs, axis=-1)
        if standardize:
            state_mean, var = self.state_moments()
            state_std = np.sqrt(var)
            state_trajs = (state_trajs - state_mean) / state_std

        return state_trajs

    def get_state_dim_names(self):
        dim_names = []
        for obs_name in self.state_obs:
            obs_dim = self.obs_dims[obs_name]
            dim_names += [f"{obs_name}:{i}" for i in range(obs_dim)]
        return dim_names

    def get_obs_from_vector(self, vector: np.ndarray, obs_names: tuple[str, ...] = None) -> dict[str, np.ndarray]:
        """Extract the observation values from a vector given the ordered observation names."""
        if obs_names is None:
            obs_names = self.state_obs
        else:
            for name in obs_names:
                assert name in self.recordings.keys(), f"Observation {name} not in recordings {self.recordings.keys()}"
        obs_dims = [self.obs_dims[obs] for obs in obs_names]
        assert vector.shape[-1] == sum(obs_dims), (
            f"Vector dim {vector.shape[-1]} differs from expected dimension {sum(obs_dims)}"
        )
        obs_idx = [0] + [sum(obs_dims[:i]) for i in range(1, len(obs_dims))]
        obs_values = {
            obs_name: vector[..., obs_idx[i] : obs_idx[i] + obs_dims[i]] for i, obs_name in enumerate(obs_names)
        }
        return obs_values

    def save_to_file(self, file_path: Path):
        # Store representations and groups without serializing
        if len(self.obs_representations) > 0:
            self._obs_rep_irreps = {}
            self._obs_rep_names = {}
            self._obs_rep_Q = {}
            for k, rep in self.obs_representations.items():
                self._obs_rep_irreps[k] = rep.irreps if rep is not None else None
                self._obs_rep_names[k] = rep.name if rep is not None else None
                self._obs_rep_Q[k] = rep.change_of_basis if rep is not None else None
            group = self.obs_representations[self.state_obs[0]].group
            self._group_keys = group._keys
            self._group_name = group.__class__.__name__
            # Remove non-serializable objects
            del self.obs_representations
            self.dynamics_parameters.pop("group", None)

        with file_path.with_suffix(".pkl").open("wb") as file:
            self._path = file_path.with_suffix(".pkl").absolute()
            pickle.dump(self, file, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load_from_file(
        file_path: Path, only_metadata=False, obs_names: Optional[Iterable[str]] = None, ignore_other_obs: bool = True
    ) -> "DynamicsRecording":
        """Load a DynamicsRecording object from a file.

        Args:
            file_path: Path to the file containing the DynamicsRecording object.
            only_metadata: If True, the recordings are not loaded, only the metadata.
            obs_names: List of observation names to load. If None, all observations are loaded.
            ignore_other_obs: If True, observations not in `obs_names` are ignored.

        Returns:
            A DynamicsRecording object containing the loaded data.
        """
        with file_path.with_suffix(".pkl").open("rb") as file:
            data = pickle.load(file)
            if only_metadata:
                del data.recordings
            else:
                data_obs_names = list(data.recordings.keys())
                if obs_names is not None:
                    data.state_obs = tuple(obs_names)
                    if ignore_other_obs:
                        irrelevant_obs = [k for k in data_obs_names if k not in obs_names and obs_names is not None]
                        for k in irrelevant_obs:
                            del data.recordings[k]

            if hasattr(data, "_group_name"):
                group = groups_dict[data._group_name]._generator(**data._group_keys)  # Instanciate symmetry group
                data.dynamics_parameters["group"] = group
                data.obs_representations = {}
                for obs_name in data._obs_rep_irreps.keys():
                    irreps_ids = data._obs_rep_irreps[obs_name]
                    rep_name = data._obs_rep_names[obs_name]
                    rep_Q = data._obs_rep_Q[obs_name]
                    if rep_name is None:
                        data.obs_representations[obs_name] = None
                    elif rep_name in group.representations:
                        data.obs_representations[obs_name] = group.representations[rep_name]
                    else:
                        data.obs_representations[obs_name] = Representation(
                            group, name=rep_name, irreps=irreps_ids, change_of_basis=rep_Q
                        )
                    group.representations[rep_name] = data.obs_representations[obs_name]
        data.__post_init__()
        return data

    @staticmethod
    def load_data_generator(
        dynamics_recordings: list["DynamicsRecording"],
        frames_per_step: int = 1,
        prediction_horizon: Union[int, float] = 1,
        state_obs: Optional[list[str]] = None,
        action_obs: Optional[list[str]] = None,
    ):
        """Generator that yields observation samples of length `n_frames_per_state` from the Markov Dynamics recordings.

        Args:
            recordings (list[DynamicsRecording]): List of DynamicsRecordings.
            frames_per_step: Number of frames to compose a single observation sample at time `t`. E.g. if `f` is
            provided
            the state samples will be of shape [f, obs_dim].
            prediction_horizon (int, float): Number of future time steps to include in the next time samples.
                E.g: if `n` is an integer the samples will be of shape [n, frames_per_state, obs_dim]
                If `n` is a float, then the samples will be of shape [int(n*traj_length), frames_per_state, obs_dim]
            state_obs: Ordered list of observations names composing the state space.
            action_obs: Ordered list of observations names composing the action space.

        Returns:
            A dictionary containing the observations of shape (time_horizon, frames_per_step, obs_dim)
        """
        for file_data in dynamics_recordings:
            recordings = file_data.recordings
            relevant_obs = set(file_data.state_obs).union(set(file_data.action_obs))
            if state_obs is not None:
                relevant_obs = set(state_obs)
            if action_obs is not None:
                relevant_obs = relevant_obs.union(set(action_obs))

            # Get any observation list of trajectories and count the number of trajectories
            # We assume all observations have the same number of trajectories
            n_trajs = len(next(iter(recordings.values())))

            # Since we assume trajectories can have different lengths, we iterate over each trajectory
            # and generate samples of length `n_frames_per_state` from each trajectory.
            for traj_id in range(n_trajs):
                traj_length = next(iter(recordings.values()))[traj_id].shape[0]
                if isinstance(prediction_horizon, float):
                    steps_in_pred_horizon = floor((prediction_horizon * traj_length) // frames_per_step) - 1
                else:
                    steps_in_pred_horizon = prediction_horizon
                assert steps_in_pred_horizon > 0, f"Invalid prediction horizon {steps_in_pred_horizon}"

                remnant = traj_length % frames_per_step
                frames_in_pred_horizon = steps_in_pred_horizon * frames_per_step
                # Iterate over the frames of the trajectory
                for frame in range(traj_length - frames_per_step):
                    # Collect the next steps until the end of the trajectory. If the prediction horizon is larger than
                    # the remaining steps (prediction outside trajectory length), then we continue to the next traj.
                    # This is better computationally for avoiding copying while batch processing data later.
                    if frame + frames_per_step + frames_in_pred_horizon > (traj_length - remnant):
                        continue
                    sample = {}
                    for obs_name, trajs in recordings.items():
                        if obs_name not in relevant_obs:  # Do not process unrequested observations
                            continue
                        num_steps = (frames_in_pred_horizon // frames_per_step) + 1
                        # Compute the indices for the start and end of each "step" in the time horizon
                        start_indices = np.arange(0, num_steps) * frames_per_step + frame
                        end_indices = start_indices + frames_per_step
                        # Use these indices to slice the relevant portion of the trajectory
                        obs_time_horizon = trajs[traj_id][start_indices[0] : end_indices[-1]]
                        # Reshape the slice to have the desired shape (time, frames_per_step, obs_dim)
                        obs_dim = file_data.obs_dims[obs_name]
                        obs_time_horizon = obs_time_horizon.reshape((num_steps, frames_per_step, obs_dim))

                        # Test no copy is being made (too costly to do at runtime)
                        # assert np.shares_memory(obs_time_horizon, trajs[traj_id])
                        assert len(obs_time_horizon) == steps_in_pred_horizon + 1, (
                            f"{len(obs_time_horizon)} != {steps_in_pred_horizon + 1}"
                        )
                        sample[obs_name] = obs_time_horizon
                    # print(frame)
                    yield sample

    @staticmethod
    def map_state_next_state(
        sample: dict,
        state_observations: List[str],
        state_mean: Optional[np.ndarray] = None,
        state_std: Optional[np.ndarray] = None,
    ) -> dict:
        """Map composing multiple frames of observations into a flat vectors `state` and `next_state` samples.

        This method constructs the state `s_t` and history of nex steps `s_t+1` of the Markov Process.
        The state is defined as a set of observations within a window of fps=`frames_per_state`.
        E.g.: Consider the state is defined by the observations [m=momentum, p=position] at `fps` consecutive frames.
            Then the state at time `t` is defined as `s_t = [m_f, m_f+1,..., m_f+fps, p_f, p_f+1, ..., p_f+fps]`.
            Where we use f to denote frame in time to make the distinction from the time index `t` of the Markov
            Process.
            Then, the next state is defined as `s_t+1 = [m_f+fps,..., m_fps+fps, p_f+fps, ..., p_f+fps+fps]`.

        Args:
            sample (dict): Dictionary containing the observations of the system of shape [state_time, f].
            state_observations: Ordered list of observations names composing the state space.

        Returns:
            A dictionary containing the MDP state `s_t` and the next_state/s `[s_t+1, s_t+2, ..., s_t+pred_horizon]`.
        """
        batch_size = len(sample[f"{state_observations[0]}"])
        time_horizon = len(sample[f"{state_observations[0]}"][0])
        # Flatten observations a_t = [a_f, a_f+1, af+2, ..., a_f+F] s.t. a_t in R^{F * dim(a)}, a_f in R^{dim(a)}
        state_obs = [sample[m] for m in state_observations]
        # Define the state at time t and the states at time [t+1, t+pred_horizon]
        state_traj = np.concatenate(state_obs, axis=-1).reshape(batch_size, time_horizon, -1)
        if state_mean is not None and state_std is not None:
            state_traj = (state_traj - state_mean) / state_std
        return dict(state=state_traj[:, 0], next_state=state_traj[:, 1:])

    @staticmethod
    def map_state_action_state(sample, state_observations: List[str], action_observations: List[str]) -> dict:
        """Map composing multiple observations to single state, action, next_state samples."""
        flat_sample = DynamicsRecording.map_state_next_state(sample, state_observations)
        # Reuse the same function to for flattening action and next_action
        action_sample = DynamicsRecording.map_state_next_state(sample, action_observations)
        action_sample["action"] = action_sample.pop("state")
        action_sample["next_action"] = action_sample.pop("next_state")
        flat_sample.update(action_sample)
        return flat_sample


def estimate_dataset_size(
    recordings: list[DynamicsRecording], prediction_horizon: Union[int, float] = 1, frames_per_step: int = 1
):
    num_trajs = 0
    num_samples = 0
    steps_pred_horizon = []
    for r in recordings:
        r_num_trajs = r.info["num_traj"]
        r_traj_length = r.info["trajectory_length"]
        if isinstance(prediction_horizon, float):
            steps_in_pred_horizon = floor((prediction_horizon * r_traj_length) // frames_per_step)
        else:
            steps_in_pred_horizon = prediction_horizon
        steps_pred_horizon.append(steps_in_pred_horizon)
        frames_in_pred_horizon = steps_in_pred_horizon * frames_per_step
        samples = r_traj_length - frames_in_pred_horizon - (r_traj_length % frames_per_step) + 1
        num_samples += r_num_trajs * samples
        num_trajs += r_num_trajs
    steps_pred_horizon = np.mean(steps_pred_horizon)
    log.debug(f"Steps in prediction horizon {int(steps_pred_horizon)}")
    return num_trajs, num_samples


def reduce_dataset_size(recordings: Iterable[DynamicsRecording], train_ratio: float = 1.0):
    assert 0.0 < train_ratio <= 1.0, f"Invalid train ratio {train_ratio}"
    if train_ratio == 1.0:
        return recordings
    log.info(f"Reducing dataset size to {train_ratio * 100}%")
    # Ensure all training seeds use the same training data partitions
    from morpho_symm.utils.mysc import TemporaryNumpySeed

    with TemporaryNumpySeed(10):
        for r in recordings:
            # Decide to keep a ratio of the original trajectories
            num_trajs = r.info["num_traj"]
            if num_trajs < 10:  # Do not discard entire trajectories, but rather parts of the trajectories
                time_horizon = r.recordings[r.state_obs[0]].shape[1]  # Take the time horizon from the first observation

                # Split the trajectory into "virtual" subtrajectories, and discard some of them
                idx = np.arange(time_horizon)
                n_partitions = math.ceil((1 / (1 - train_ratio)))
                n_partitions = n_partitions * 10 if n_partitions < 10 else n_partitions
                # Ensure partitions of the same time duration
                partitions_idx = np.split(idx[: -(time_horizon % n_partitions)], indices_or_sections=n_partitions)
                partition_length = partitions_idx[0].shape[0]
                n_partitions_to_keep = math.ceil(n_partitions * train_ratio)
                partitions_to_keep = np.random.choice(range(n_partitions), size=n_partitions_to_keep, replace=False)
                print(partitions_to_keep)
                partitions_to_keep = [partitions_idx[i] for i in partitions_to_keep]

                ratio_of_samples_removed = (time_horizon - (len(partitions_to_keep) * partition_length)) / time_horizon
                assert ratio_of_samples_removed - (1 - train_ratio) < 0.05, (
                    f"Requested to remove {(1 - train_ratio) * 100}% of the samples, "
                    f"but removed {ratio_of_samples_removed * 100}%"
                )

                new_recordings = {}
                for obs_name, obs in r.recordings.items():
                    new_obs_trajs = []
                    for part_time_idx in partitions_to_keep:
                        new_obs_trajs.append(obs[:, part_time_idx])
                    new_recordings[obs_name] = np.concatenate(new_obs_trajs, axis=0)
                r.recordings = new_recordings
                r.info["num_traj"] = len(partitions_to_keep)
                r.info["trajectory_length"] = partition_length
            else:  # Discard entire trajectories
                # Sample int(num_trajs * train_ratio) trajectories from the original recordings
                num_trajs_to_keep = math.ceil(num_trajs * train_ratio)
                idx = range(num_trajs)
                idx_to_keep = np.random.choice(idx, size=num_trajs_to_keep, replace=False)
                # Keep only the selected trajectories
                new_recordings = {k: v[idx_to_keep] for k, v in r.recordings.items()}
                r.recordings = new_recordings


def split_train_val_test(
    dyn_recording: DynamicsRecording,
    partition_sizes=(0.70, 0.15, 0.15),
    split_dimension: str = "auto",  # 'time' or 'trajectory'
) -> [DynamicsRecording, DynamicsRecording, DynamicsRecording]:
    """Split the recordings into training, validation and test sets.

    Args:
        dyn_recording: (DynamicsRecording): The recordings to split.
        partition_sizes: (tuple): The sizes of the training, validation and test sets.
        split_dimension: (str): The dimension to split the data. Can be 'time' or 'trajectory' or 'auto'. In the case
         of auto, the function will decide the best way to split the data based on the shape of the data.

    Returns:
        (DynamicsRecording, DynamicsRecording, DynamicsRecording): The training, validation and test sets.
    """
    assert np.isclose(np.sum(partition_sizes), 1.0), f"Invalid partition sizes {partition_sizes}"
    partitions_names = ["train", "val", "test"]

    log.info(f"Partitioning {dyn_recording.description} into train/val/test of sizes {partition_sizes}[%]")
    # Ensure all training seeds use the same training data partitions
    from morpho_symm.utils.mysc import TemporaryNumpySeed

    with TemporaryNumpySeed(10):  # Ensure deterministic behavior
        # Decide to keep a ratio of the original trajectories
        state_traj = dyn_recording.get_state_trajs()
        assert state_traj.ndim == 3, f"Expectec (traj, time, state_dim) but got {state_traj.shape}"
        num_trajs, time_horizon, state_dim = state_traj.shape

        if split_dimension == "auto":
            split_time = time_horizon > num_trajs
        elif split_dimension == "time":
            split_time = True
        elif split_dimension == "trajectory":
            split_time = False
        else:
            raise ValueError(f"Invalid split dimension {split_dimension} expected 'time', 'trajectory' or 'auto'")

        if split_time:  # Do not discard entire trajectories, but rather parts of the trajectories
            num_samples = time_horizon
            min_idx = 0
            partitions_sample_idx = {partition: None for partition in partitions_names}
            for partition_name, ratio in zip(partitions_names, partition_sizes):
                max_idx = min_idx + int(num_samples * ratio)
                partitions_sample_idx[partition_name] = list(range(min_idx, max_idx))
                min_idx = min_idx + int(num_samples * ratio)

            # TODO: Avoid deep copying the data itself.
            partitions_recordings = {partition: copy.deepcopy(dyn_recording) for partition in partitions_names}
            for partition_name, sample_idx in partitions_sample_idx.items():
                part_num_samples = len(sample_idx)
                partitions_recordings[partition_name].info["trajectory_length"] = part_num_samples
                partitions_recordings[partition_name].recordings = dict()
                for obs_name in dyn_recording.recordings.keys():
                    if len(dyn_recording.recordings[obs_name].shape) == 3:
                        data = dyn_recording.recordings[obs_name][:, sample_idx]
                    elif len(dyn_recording.recordings[obs_name].shape) == 2:
                        data = dyn_recording.recordings[obs_name][sample_idx]
                    else:
                        raise RuntimeError(f"Invalid shape {dyn_recording.recordings[obs_name].shape} of {obs_name}")
                    partitions_recordings[partition_name].recordings[obs_name] = data

        else:  # Select train/val/test from individual trajectories
            num_samples = num_trajs
            min_idx = 0
            partitions_sample_idx = {partition: None for partition in partitions_names}
            for partition_name, ratio in zip(partitions_names, partition_sizes):
                max_idx = min_idx + int(num_samples * ratio)
                partitions_sample_idx[partition_name] = list(range(min_idx, max_idx))
                min_idx = min_idx + int(num_samples * ratio)

            partitions_recordings = {partition: copy.deepcopy(dyn_recording) for partition in partitions_names}
            for partition_name, sample_idx in partitions_sample_idx.items():
                part_num_samples = len(sample_idx)
                partitions_recordings[partition_name].info["num_traj"] = part_num_samples
                partitions_recordings[partition_name].recordings = dict()
                for obs_name in dyn_recording.recordings.keys():
                    data = dyn_recording.recordings[obs_name][sample_idx]
                    partitions_recordings[partition_name].recordings[obs_name] = data

        return partitions_recordings["train"], partitions_recordings["val"], partitions_recordings["test"]


def get_dynamics_dataset(
    train_shards: list[Path],
    test_shards: Optional[list[Path]] = None,
    val_shards: Optional[list[Path]] = None,
    frames_per_step: int = 1,
    train_ratio: float = 1.0,
    train_pred_horizon: Union[int, float] = 1,
    eval_pred_horizon: Union[int, float] = 10,
    test_pred_horizon: Union[int, float] = 10,
    state_obs: Optional[tuple[str]] = None,
    action_obs: Optional[tuple[str]] = None,
    hard_test_augment: bool = False,
    hard_val_augment: bool = False,
) -> tuple[list[IterableDataset], DynamicsRecording]:
    """Load Markov Dynamics recordings from a list of files and return a train, test and validation dataset."""
    # TODO: ensure all shards come from the same dynamical system
    metadata: DynamicsRecording = DynamicsRecording.load_from_file(train_shards[0])
    test_shards = [] if test_shards is None else test_shards
    val_shards = [] if val_shards is None else val_shards

    if len(test_shards) > 1:
        from morpho_symm.utils.mysc import compare_dictionaries

        test_metadata = DynamicsRecording.load_from_file(test_shards[0], only_metadata=True)
        dyn_params_diff = compare_dictionaries(metadata.dynamics_parameters, test_metadata.dynamics_parameters)
        assert len(dyn_params_diff) == 0, "Different dynamical systems loaded in train/test sets"

    # Define the relevant observations of the recording to load.
    state_obs = state_obs if state_obs is not None else metadata.state_obs
    action_obs = action_obs if action_obs is not None else metadata.action_obs
    relevant_obs = set(state_obs).union(set(action_obs))
    features = {}
    assert len(relevant_obs) > 0, "Provide the names of the observations to be included in state (and action)"
    for obs_name in relevant_obs:
        assert obs_name in metadata.recordings.keys(), f"Observation {obs_name} not found in recordings"
    for obs_name in relevant_obs:
        features[obs_name] = datasets.Array2D(shape=(frames_per_step, metadata.obs_dims[obs_name]), dtype="float32")

    part_datasets = []
    for partition, partition_shards in zip(["train", "test", "val"], [train_shards, test_shards, val_shards]):
        if len(partition_shards) == 0:
            part_datasets.append(None)
            continue

        recordings = [DynamicsRecording.load_from_file(f, obs_names=relevant_obs) for f in partition_shards]
        a = recordings[0]
        mean, var = a.state_moments()
        if partition == "train":
            pred_horizon = train_pred_horizon
            if train_ratio < 1.0:
                reduce_dataset_size(recordings, train_ratio)
        elif partition == "val":
            pred_horizon = eval_pred_horizon
        else:
            pred_horizon = test_pred_horizon

        num_trajs, num_samples = estimate_dataset_size(recordings, pred_horizon, frames_per_step)
        dataset = IterableDataset.from_generator(
            DynamicsRecording.load_data_generator,
            features=Features(features),
            gen_kwargs=dict(
                dynamics_recordings=recordings,
                frames_per_step=frames_per_step,
                prediction_horizon=pred_horizon,
                state_obs=tuple(state_obs),
                action_obs=tuple(action_obs),
            ),
        )

        log.debug(
            f"[Dataset {partition} - Trajs:{num_trajs} - Samples: {num_samples} - "
            f"Frames per sample : {frames_per_step}]-----------------------------"
        )

        dataset.info.dataset_size = num_samples
        dataset.info.dataset_name = f"[{partition}] Linear dynamics"
        dataset.info.description = metadata.description
        # dataset = partition
        part_datasets.append(dataset)

    metadata.state_obs = state_obs
    metadata.action_obs = action_obs
    return part_datasets, metadata


def get_train_test_val_file_paths(data_path: Path):
    """Search in folder for files ending in train/test/val.pkl and return a list of paths to each file."""
    train_data, test_data, val_data = [], [], []
    for file_path in data_path.iterdir():
        if file_path.name.endswith("train.pkl"):
            train_data.append(file_path)
        elif file_path.name.endswith("test.pkl"):
            test_data.append(file_path)
        elif file_path.name.endswith("val.pkl"):
            val_data.append(file_path)
    return train_data, test_data, val_data


if __name__ == "__main__":
    path_to_data = Path(__file__).parent
    assert path_to_data.exists(), f"Invalid Dataset path {path_to_data.absolute()}"

    # Find all dynamic systems recordings
    path_to_data /= Path("mini_cheetah") / "recordings" / "grass"
    # path_to_data = Path('/home/danfoa/Projects/koopman_robotics/data/linear_system/group=C3-dim=6/n_constraints=1/')
    path_to_dyn_sys_data = set([a.parent for a in list(path_to_data.rglob("*test.pkl"))])
    # Select a dynamical system
    mock_path = path_to_dyn_sys_data.pop()
    # Obtain the training, testing and validation file paths containing distinct trajectories of motion.
    train_data, test_data, val_data = get_train_test_val_file_paths(mock_path)
    # Obtain hugging face Iterable datasets instances
    pred_horizon = 0.1
    frames_per_state = 10
    (train_dataset, test_dataset, val_dataset), metadata = get_dynamics_dataset(
        train_shards=train_data,
        test_shards=test_data,
        val_shards=val_data,
        frames_per_step=frames_per_state,
        train_pred_horizon=pred_horizon,
        eval_pred_horizon=0.5,
    )

    # test_flat = map_state_next_state(sample, metadata.state_observations)
    # Test the map
    torch_dataset = test_dataset.with_format("torch").shuffle()
    torch_dataset = torch_dataset.map(map_state_next_state, batched=True, fn_kwargs={"state_observations": ["state"]})

    # Test errors while looping.
    for i, s in enumerate(torch_dataset):
        pass
        # log.debug(s)
        # if i > 1000:
        #     break

    log.debug(i)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 13/5/22
# @Author  : Daniel Ordonez
# @email   : daniels.ordonez@gmail.com
import glob
import io
import pathlib
import sys
import time

import numpy as np
import pandas as pd
import PIL
import scipy
import sklearn
import torch
import torch.nn.functional as F
from omegaconf import DictConfig
from sklearn.metrics import jaccard_score, precision_score

from morpho_symm.utils.robot_utils import load_symmetric_system

try:
    deep_contact_estimator_path = pathlib.Path(__file__).parent.absolute()
    assert deep_contact_estimator_path.exists()
    sys.path.append(str(deep_contact_estimator_path / "deep-contact-estimator/utils"))
    sys.path.append(str(deep_contact_estimator_path / "deep-contact-estimator/src"))
    from data_handler import contact_dataset
except ImportError as e:
    raise ImportError(
        "Deep Contact Estimator submodule not initialized, run `git submodule update --init --recursive --progress"
    ) from e

from pytorch_lightning import Trainer
from sklearn.metrics import confusion_matrix
from torch.utils.data._utils.collate import default_collate
from tqdm import tqdm


class UmichContactDataset(contact_dataset):
    dataset_path = pathlib.Path(__file__).parent
    leg_names = ["RF", "LF", "RH", "LH"]
    states_names = [
        "-",
        "LH",
        "RH",
        "RH-LH",
        "LF",
        "LF-LH",
        "LF-RH",
        "LF-RH-LH",
        "RF",
        "RF-LH",
        "RF-RH",
        "RF-RH-LH",
        "RF-LF",
        "RF-LF-LH",
        "RF-LF-RH",
        "RF-LF-RH-LH",
    ]

    def __init__(
        self,
        data_name,
        label_name,
        window_size,
        robot_cfg,
        train_ratio=0.7,
        val_ratio=0.15,
        loss_class_weights=None,
        use_class_imbalance_w=False,
        device="cuda",
        augment=False,
        debug=False,
        partition="training",
    ):
        # Sub folder in dataset folder containing the mat/*.mat and numpy/*.npy
        self.partition = partition
        self.data_path, self.label_path = self.get_full_paths(
            data_name, label_name, train_ratio=train_ratio, val_ratio=val_ratio
        )
        self.robot_cfg = robot_cfg

        trials = 5
        data, label = None, None
        while data is None and trials > 0:
            try:
                data = np.load(str(self.data_path))
                label = np.load(str(self.label_path))
            except:
                trials -= 1
                time.sleep(np.random.random())

        self.num_data = data.shape[0] - window_size + 1
        self.window_size = window_size
        self.data = torch.from_numpy(data).type("torch.FloatTensor").to(device)
        self.label = torch.from_numpy(label).type("torch.LongTensor").to(device)
        # ----
        self.device = device

        self.contact_state_freq = slabelelf.get_class_frequency()

        self.rep_in, self.rep_out = self.get_in_out_symmetry_groups_reps(robot_cfg)
        self.augment = augment
        self.n_contact_states = 16

        self.hin = torch.tensor(self.rep_in.G.discrete_generators[0].todense(), device=device)
        self.hout = torch.tensor(self.rep_out.G.discrete_generators[0].todense(), device=device)

        if use_class_imbalance_w:
            self.class_weights = 1 - self.contact_state_freq if loss_class_weights is None else loss_class_weights
        else:
            self.class_weights = None
        self.loss_fn = torch.nn.CrossEntropyLoss(weight=self.class_weights)

        if debug:
            self.plot_statistics()

    def collate_fn(self, batch):
        collated_batch = default_collate(batch)

        if self.augment and np.random.rand() > 0.5:
            # (Batch, Window size, features)
            x_batch = collated_batch["data"]
            y_batch = F.one_hot(collated_batch["label"], num_classes=self.n_contact_states).to(x_batch.dtype)
            g_x_batch = torch.matmul(x_batch.unsqueeze(1), self.hin.unsqueeze(0).to(x_batch.dtype)).squeeze()
            g_y_batch = torch.matmul(y_batch.unsqueeze(1), self.hout.unsqueeze(0).to(x_batch.dtype)).squeeze()
            # Convert back to numerical class label.
            _, g_y = torch.max(g_y_batch, dim=1)
            # return {"data": g_x_batch, "label": g_y}
            return g_x_batch, g_y
        else:
            # return collated_batch
            return collated_batch["data"], collated_batch["label"]

    def get_class_frequency(self):
        classes, counts = torch.unique(self.label, return_counts=True, sorted=True)
        return counts / len(self)

    @torch.no_grad()
    def compute_metrics(self, output: torch.Tensor, gt_label: torch.Tensor) -> dict:
        _, prediction = torch.max(output, dim=-1)

        bin_pred = self.decimal2binary(prediction)
        bin_gt = self.decimal2binary(gt_label)

        acc_per_leg = (bin_pred == bin_gt).sum(axis=0) / output.shape[0]
        acc = (prediction == gt_label).sum() / output.shape[0]
        acc_per_leg_avg = torch.sum(acc_per_leg) / 4.0

        acc_dir = {f"{leg}/acc": v for leg, v in zip(self.leg_names, acc_per_leg)}
        metrics = {"contact_state/acc": acc, "legs_avg/acc": acc_per_leg_avg}
        metrics.update(acc_dir)
        return metrics

    def test_metrics(self, y_pred, y_gt, trainer: Trainer, model, log_imgs=False, prefix="test_"):
        model.eval()

        # acc_dir = self.compute_metrics(y_pred, y_gt)
        _, prediction = torch.max(y_pred, dim=-1)
        bin_pred_arr = self.decimal2binary(prediction).detach().cpu().numpy()
        bin_gt_arr = self.decimal2binary(y_gt).detach().cpu().numpy()
        pred_arr = prediction.detach().cpu().numpy()
        gt_arr = y_gt.detach().cpu().numpy()

        # test_acc, acc_per_leg, bin_pred_arr, bin_gt_arr, pred_arr, gt_arr = self.compute_accuracy(dataloader, model)
        precision_of_class, precision_of_legs, precision_of_all_legs = self.compute_precision(
            bin_pred_arr, bin_gt_arr, pred_arr, gt_arr
        )
        f1score_of_class = sklearn.metrics.f1_score(gt_arr, pred_arr, average="weighted")

        jaccard_of_class, jaccard_of_legs, jaccard_of_all_legs = self.compute_jaccard(
            bin_pred_arr, bin_gt_arr, pred_arr, gt_arr
        )
        confusion_mat, rates = self.compute_confusion_mat(bin_pred_arr, bin_gt_arr, pred_arr, gt_arr)

        TPR = [rates[f"{k}/TP"] / np.sum(bin_gt_arr[:, i] == 1) for i, k in enumerate(self.leg_names)]
        TNR = [rates[f"{k}/TN"] / np.sum(bin_gt_arr[:, i] == 0) for i, k in enumerate(self.leg_names)]
        FNR = [1 - tpr for tpr in TPR]
        FPR = [1 - tpr for tpr in TNR]

        balanced_acc_of_legs = [(tpr + tnr) / 2 for tpr, tnr in zip(TPR, TNR)]
        recall_of_legs = [rates[f"{k}/TP"] / (rates[f"{k}/TP"] + rates[f"{k}/FP"]) for k in self.leg_names]
        f1score_of_legs = [2 * (p * r) / (r + p) for p, r in zip(precision_of_legs, recall_of_legs)]
        pred_pos_cond_rate_of_legs = [
            (rates[f"{k}/TP"] + rates[f"{k}/FP"])
            / (rates[f"{k}/TP"] + rates[f"{k}/FP"] + rates[f"{k}/TN"] + rates[f"{k}/FN"])
            for k in self.leg_names
        ]

        individual_state_metrics = {}
        if prefix == "test_":
            for state_id, state_name in enumerate(self.states_names):
                precision, recall, f1, support = sklearn.metrics.precision_recall_fscore_support(
                    y_true=gt_arr, y_pred=pred_arr, labels=[state_id]
                )
                individual_state_metrics[f"contact_state/{state_name}/precision"] = float(precision)
                individual_state_metrics[f"contact_state/{state_name}/recall"] = float(recall)
                individual_state_metrics[f"contact_state/{state_name}/f1"] = float(f1)
                individual_state_metrics[f"contact_state/{state_name}/support"] = int(np.sum(gt_arr == state_id))

        if log_imgs:
            import matplotlib.pyplot as plt
            import seaborn as sns

            def log_cm_img(label, classes, figsize=(4, 3), annot=True):
                # If Trainer is present log images
                df_cfm = pd.DataFrame(confusion_mat[label], index=classes, columns=classes)
                fig = plt.figure(figsize=figsize, dpi=80)
                sns.heatmap(df_cfm, annot=annot, fmt=".2f", vmin=0, vmax=1)
                fig.tight_layout()
                # plt.show()
                buf = io.BytesIO()
                plt.savefig(buf, format="jpeg")
                buf.seek(0)
                img = PIL.Image.open(buf)
                plt.close(fig)

                if trainer.logger:
                    from torchvision.transforms import ToTensor

                    tensorboard = trainer.logger.experiment
                    tensorboard.add_image(label, ToTensor()(img))

            log_cm_img(label="cm/contact_state", classes=self.states_names, annot=False, figsize=(8, 6))
            for key, cm in confusion_mat.items():
                if "contact_state" in key:
                    continue
                log_cm_img(label=key, classes=[0, 1], annot=True)

        precision_dir = {f"{leg}/precision": v for leg, v in zip(self.leg_names, precision_of_legs)}
        jaccard_dir = {f"{leg}/jaccard": v for leg, v in zip(self.leg_names, jaccard_of_legs)}
        recall_dir = {f"{leg}/recall": v for leg, v in zip(self.leg_names, recall_of_legs)}
        f1_score_dir = {f"{leg}/f1": v for leg, v in zip(self.leg_names, f1score_of_legs)}
        balanced_acc_dir = {f"{leg}/balanced_acc": a for leg, a in zip(self.leg_names, balanced_acc_of_legs)}

        metrics = {
            "contact_state/f1": f1score_of_class,
            "contact_state/precision": precision_of_class,
            "contact_state/jaccard": jaccard_of_class,
            "legs_avg/precision": precision_of_all_legs,
            "legs_avg/jaccard": jaccard_of_all_legs,
            "legs_avg/f1": np.sum(f1score_of_legs) / 4.0,
            "legs_avg/recall": np.sum(recall_of_legs) / 4.0,
            "legs_avg/balanced_acc": np.sum(balanced_acc_of_legs) / 4.0,
            "legs_avg/specificity": np.sum(TNR) / 4.0,
            "legs_avg/FPR": np.sum(FPR) / 4.0,
            "legs_avg/FNR": np.sum(FNR) / 4.0,
            "legs_avg/pred_pos_cond_rate": np.sum(pred_pos_cond_rate_of_legs) / 4.0,
        }

        metrics.update(precision_dir)
        metrics.update(jaccard_dir)
        metrics.update(recall_dir)
        metrics.update(f1_score_dir)
        metrics.update(balanced_acc_dir)
        metrics.update(individual_state_metrics)

        model.log_metrics(metrics, suffix=prefix)
        model.train()

    def decimal2binary(self, x):
        mask = 2 ** torch.arange(4 - 1, -1, -1).to(self.device, x.dtype)
        return x.unsqueeze(-1).bitwise_and(mask).ne(0).byte()

    @staticmethod
    def get_in_out_symmetry_groups_reps(robot_cfg: DictConfig):
        from morpho_symm.groups.SparseRepresentation import SparseRep

        robot, rep_E3, rep_QJ = load_symmetric_system(robot_cfg)
        G_class = class_from_name(
            "groups.SymmetryGroups", robot_cfg.G if robot_cfg.gens_ids is None else robot_cfg.G_sub
        )

        # Configure input x and output y representations
        # x = [qj, dqj, a, w, pf, vf] - a: linear IMU acc, w: angular IMU velocity, pf:feet positions, vf:feet velocity
        rep_a = rep_E3.set_pseudovector(False)
        rep_w = rep_E3.set_pseudovector(True)

        # Configure pf, vf ∈ R^12  representations composed of reflections and permutations
        n_legs = 4
        rep_legs_reflected = n_legs * rep_E3.set_pseudovector(
            False
        )  # Same representation as the hips ref frames are collinear with base.
        G_legs_reflected = rep_legs_reflected.G
        g_q_perm = abs(rep_QJ.G.discrete_generators[0])  # Permutation swapping legs.
        G_pf = G_class(generators=g_q_perm @ G_legs_reflected.discrete_generators[0])
        rep_pf = SparseRep(G_pf)
        rep_vf = SparseRep(G_pf)

        # x   = [ qj   ,  dqj   ,   a   ,   w   ,   pf   ,   vf  ]
        rep_x = rep_QJ + rep_QJ + rep_a + rep_w + rep_pf + rep_vf
        # y : 16 dimensional contact state with following symmetry. See paper abstract.
        g_y = C2.oneline2matrix(oneline_notation=[0, 2, 1, 3, 8, 10, 9, 11, 4, 6, 5, 7, 12, 14, 13, 15])
        G_y = C2(g_y)
        rep_y = SparseRep(G_y)

        rep_data_in, rep_data_out = rep_x, rep_y
        return rep_data_in, rep_data_out

    def get_full_paths(
        self, data_name, label_name, train_ratio: float = 0.7, val_ratio: float = 0.7
    ) -> (pathlib.Path, pathlib.Path):
        folder_path = pathlib.Path(self.dataset_path / f"{self.partition}/numpy_train_ratio={train_ratio:.3f}")
        training_mat_path = pathlib.Path(self.dataset_path / f"{self.partition}/mat")
        test_mat_path = pathlib.Path(self.dataset_path / f"{self.partition}/mat_test")

        data_path = folder_path.joinpath(data_name)
        label_path = folder_path.joinpath(label_name)
        print(f"Contact Dataset path: \n\t- Data: {data_path} \n\t- Labels: {label_path}")

        if not data_path.exists():  # Data is not there generate it.
            folder_path.mkdir(exist_ok=True)
            print(f"Generating dataset and saving it to: {folder_path}")
            UmichContactDataset.mat2numpy_split(
                train_val_data_path=training_mat_path,
                test_data_path=test_mat_path,
                save_path=folder_path,
                train_ratio=train_ratio,
                val_ratio=val_ratio,
            )
            print(f"Created dataset partition and saved to: {folder_path}")
            assert data_path.exists(), f"Failed to create partition on {data_path}"

        return data_path, label_path

    def compute_accuracy(self, dataloader, model):
        # compute accuracy in batch

        num_correct = 0
        num_data = 0
        correct_per_leg = np.zeros(4)
        bin_pred_arr = np.zeros((0, 4))
        bin_gt_arr = np.zeros((0, 4))
        pred_arr = np.zeros((0))
        gt_arr = np.zeros((0))
        with torch.no_grad():
            for input_data, gt_label in tqdm(dataloader):
                output = model(input_data)

                _, prediction = torch.max(output, 1)

                bin_pred = self.decimal2binary(prediction)
                bin_gt = self.decimal2binary(gt_label)

                bin_pred_arr = np.vstack((bin_pred_arr, bin_pred.cpu().numpy()))
                bin_gt_arr = np.vstack((bin_gt_arr, bin_gt.cpu().numpy()))

                pred_arr = np.hstack((pred_arr, prediction.cpu().numpy()))
                gt_arr = np.hstack((gt_arr, gt_label.cpu().numpy()))

                correct_per_leg += (bin_pred == bin_gt).sum(axis=0).cpu().numpy()
                num_data += input_data.size(0)
                num_correct += (prediction == gt_label).sum().item()

        return num_correct / num_data, correct_per_leg / num_data, bin_pred_arr, bin_gt_arr, pred_arr, gt_arr

    @staticmethod
    def mat2numpy_split(
        train_val_data_path: pathlib.Path,
        test_data_path: pathlib.Path,
        save_path: pathlib.Path,
        train_ratio=0.7,
        val_ratio=0.15,
    ):
        """Load data from .mat file, concatenate into numpy array, and save as train/val/test.

        Inputs:
        - data_pth: path to mat data folder
        - save_pth: path to numpy saving directory.
        - train_ratio: ratio of training data
        - val_ratio: ratio of validation data
        Data should be stored in .mat file, and contain:
        - q: joint encoder value (num_data,12)
        - qd: joint angular velocity (num_data,12)
        - p: foot position from FK (num_data,12)
        - v: foot velocity from FK (num_data,12)
        - imu_acc: linear acceleration from imu (num_data,3)
        - imu_omega: angular velocity from imu (num_data,3)
        - contacts: contact_hp_ecnn data (num_data,4)
                    contacts are stored as binary values, in the order
                    of right_front, left_front, right_hind, left_hind.
                    FRONT
                    1 0  RIGHT
                    3 2
                    BACK
                    Contact value will be treated as binary values in the
                    order of contacts[0], contacts[1], contacts[2], contacts[3]
                    and be converted to decimal value in this function.
                    Ex. [1,0,0,1] -> 9
                        [0,1,1,0] -> 6.

        - tau_est (optional): estimated control torque (num_data,12)
        - F (optional): ground reaction force

        Output:
        -
        """
        assert train_ratio + val_ratio <= 1.0, f"{(train_ratio, val_ratio)} > 1.0"

        data_path = pathlib.Path(train_val_data_path)
        test_data_path = pathlib.Path(test_data_path)
        save_path = pathlib.Path(save_path)
        assert data_path.exists(), f"Train and Val .mat files required, and not found on {data_path.absolute()}"
        assert test_data_path.exists(), f"Test .mat files required, and not found on {test_data_path.absolute()}"
        assert data_path != test_data_path

        save_path.mkdir(exist_ok=True)

        print(f"\nCreating dataset partition: train:{train_ratio}, val:{val_ratio} from {data_path}")
        (train_data, val_data), (train_label, val_label) = UmichContactDataset.load_and_split_mat_files(
            data_path=train_val_data_path, partitions_ratio=(train_ratio, val_ratio), partitions_name=("train", "val")
        )
        print(f"\nCreating dataset test from {test_data_path}")
        (test_data,), (test_label,) = UmichContactDataset.load_and_split_mat_files(
            data_path=test_data_path, partitions_ratio=(1.0,), partitions_name=("test",)
        )

        print(f"Saving data to {save_path.resolve()}")

        np.save(str(save_path.joinpath("train.npy")), train_data)
        np.save(str(save_path.joinpath("val.npy")), val_data)
        np.save(str(save_path.joinpath("test.npy")), test_data)
        np.save(str(save_path.joinpath("train_label.npy")), train_label)
        np.save(str(save_path.joinpath("val_label.npy")), val_label)
        np.save(str(save_path.joinpath("test_label.npy")), test_label)

    @staticmethod
    def load_and_split_mat_files(
        data_path: pathlib.Path, partitions_ratio=(0.85, 0.15), partitions_name=("train", "val")
    ):
        partitions_data = [None] * len(partitions_ratio)
        partitions_labels = [None] * len(partitions_ratio)
        data_files = list(data_path.glob("*.mat"))
        assert len(data_files) > 1, f"No .mat files found in {data_path.absolute()}"
        assert sum(partitions_ratio) <= 1.0, "the partitions should add up to less than 100% of the data"

        print(f"Loading data from {data_path}")
        print(f"Dataset .mat files found: {[pathlib.Path(d).name for d in data_files]}")
        # for all dataset in the folder
        all_samples = 0
        for data_name in glob.glob(str(data_path.joinpath("*"))):
            # load data
            raw_data = scipy.io.loadmat(data_name)

            contacts = raw_data["contacts"]
            q = raw_data["q"]
            p = raw_data["p"]
            qd = raw_data["qd"]
            v = raw_data["v"]
            acc = raw_data["imu_acc"]
            omega = raw_data["imu_omega"]

            # concatenate current data. First we try without GRF
            cur_data = np.concatenate((q, qd, acc, omega, p, v), axis=1)

            # convert labels from binary to decimal
            def binary2decimal(a, axis=-1):
                return np.right_shift(np.packbits(a, axis=axis), 8 - a.shape[axis]).squeeze()

            cur_label = binary2decimal(contacts).reshape((-1, 1))

            # separate data into given paritions
            num_data = np.shape(q)[0]
            all_samples += num_data
            partitions_size = [int(ratio * num_data) for ratio in partitions_ratio]

            lower_lim = 0
            for partition_id, (partition_size, partition_name) in enumerate(zip(partitions_size, partitions_name)):
                if partitions_data[partition_id] is None:
                    partitions_data[partition_id] = []
                    partitions_labels[partition_id] = []
                partitions_data[partition_id].append(cur_data[lower_lim : lower_lim + partition_size, :])
                partitions_labels[partition_id].append(cur_label[lower_lim : lower_lim + partition_size, :])
                lower_lim += partition_size

        partitions_data = [np.vstack(data_list) for data_list in partitions_data]
        partitions_labels = [
            np.vstack(label_list).reshape(
                -1,
            )
            for label_list in partitions_labels
        ]

        for name, ratio, data in zip(partitions_name, partitions_ratio, partitions_data):
            # assert ratio * all_samples == data.shape[0]
            print(
                f"\t - {name} ({ratio * 100:.1f}%) = {data.shape[0]:d} samples --> "
                f"{data.shape[0] / all_samples * 100:.1f}%"
            )
        return partitions_data, partitions_labels

    def compute_confusion_mat(self, bin_contact_pred_arr, bin_contact_gt_arr, pred_state, gt_state):
        confusion_mat = {}

        confusion_mat["cm/contact_state"] = confusion_matrix(
            pred_state, gt_state, labels=list(range(self.n_contact_states)), normalize="true"
        )
        for leg_id, leg_name in enumerate(self.leg_names):
            confusion_mat[f"cm/leg_{leg_name}"] = confusion_matrix(
                bin_contact_gt_arr[:, leg_id], bin_contact_pred_arr[:, leg_id], labels=[0, 1], normalize="true"
            )

        # false negative and false postivie rate
        # false negative = FN/P; false positive = FP/N
        rates = {}
        for key, cm in confusion_mat.items():
            if "contact_state" in key:
                continue
            tn, fp, fn, tp = cm.ravel()
            leg_name = key.split("_")[1]
            rates[f"{leg_name}/TP"] = tp
            rates[f"{leg_name}/FN"] = fn
            rates[f"{leg_name}/FP"] = fp
            rates[f"{leg_name}/TN"] = tn

        return confusion_mat, rates

    @staticmethod
    def compute_precision(bin_pred_arr, bin_gt_arr, pred_arr, gt_arr):
        precision_of_class = precision_score(gt_arr, pred_arr, average="weighted")
        precision_of_all_legs = precision_score(bin_gt_arr.flatten(), bin_pred_arr.flatten())
        precision_of_legs = []
        for i in range(4):
            precision_of_legs.append(precision_score(bin_gt_arr[:, i], bin_pred_arr[:, i]))

        return precision_of_class, precision_of_legs, precision_of_all_legs

    @staticmethod
    def compute_jaccard(bin_pred_arr, bin_gt_arr, pred_arr, gt_arr):
        jaccard_of_class = jaccard_score(gt_arr, pred_arr, average="weighted")
        jaccard_of_all_legs = jaccard_score(bin_gt_arr.flatten(), bin_pred_arr.flatten())
        jaccard_of_legs = []
        for i in range(4):
            jaccard_of_legs.append(jaccard_score(bin_gt_arr[:, i], bin_pred_arr[:, i]))

        return jaccard_of_class, jaccard_of_legs, jaccard_of_all_legs

    def plot_statistics(self):
        import matplotlib.pyplot as plt
        import seaborn as sns

        y = self.label.detach().cpu().numpy()
        self.decimal2binary(self.label).detach().cpu().numpy()

        df = pd.DataFrame({"contact_state": np.asarray(self.states_names)[y]})
        df["source"] = "gt"

        df2 = pd.DataFrame({"contact_state": np.asarray(self.states_names)[y]})
        df2["source"] = "pred"

        df = pd.concat((df, df2), axis=0, ignore_index=True)

        df["contact_state"] = pd.Categorical(df["contact_state"], self.states_names)
        n_colors = 1 if "source" not in df else len(np.unique(df["source"]))
        fig, ax = plt.subplots(figsize=(8, 8), dpi=150)
        sns.histplot(
            data=df,
            x="contact_state",
            hue="source" if "source" in df else None,
            color="magma",
            palette=sns.color_palette("magma_r", n_colors),
            bins=16,
            stat="probability",
            ax=ax,
            discrete=True,
            multiple="dodge",
        )
        ax.set_title(f"{self.data_path.stem} {len(self)} samples")
        ax.set(yscale="log")
        ax.tick_params(axis="x", rotation=70)
        plt.tight_layout()
        plt.show()

        # fig, axs = plt.subplots(ncols=2, nrows=2, figsize=(8, 8), dpi=150)
        # for leg_name, i, ax in zip(self.leg_names, range(4), axs.flatten()):
        #     sns.histplot(x=bin_gt[:, i], bins=2, discrete=True, shrink=.9, stat='probability', ax=ax)
        #     ax.set_title(leg_name)
        # fig.suptitle(f'{self.data_path.stem} {len(self)} samples')
        # plt.show()
        # pass



import pathlib
from collections.abc import Callable
from os import PathLike

import numpy as np
from omegaconf import OmegaConf


def load_config_hierarchy(cfg_path: PathLike):
    p = pathlib.Path(cfg_path)
    assert p.exists(), f"Cfg path {p.absolute()} does not exist"
    cfg = OmegaConf.load(p)
    if "defaults" in cfg:
        parent_cfg_names = list(cfg.defaults)
        if "_self_" in parent_cfg_names:
            parent_cfg_names.remove("_self_")
        for parent_cfg_name in parent_cfg_names:
            parent_path = p.parent / f"{parent_cfg_name}.yaml"
            parent_cfg = load_config_hierarchy(parent_path)
            cfg = OmegaConf.merge(parent_cfg, cfg)
    return cfg


class CallableDict(dict, Callable):
    def __call__(self, key):
        return self[key]


def flatten_dict(d: dict, prefix=""):
    a = {}
    for k, v in d.items():
        if isinstance(v, dict):
            a.update(flatten_dict(v, prefix=f"{k}/"))
        else:
            a[f"{prefix}{k}"] = v
    return a


def compare_dictionaries(dict1: dict, dict2: dict):
    """Recursively compare dictionaries which entries might be other dictionaries.
    Any entry in one dictionary that is not present in the other dictionary needs to be reported.

    Args:
        dict1: A tree of dictionaries.
        dict2: A tree of dictionaries.

    Returns: The dictionary containing only the keys that are different between the two dictionaries. In all levels of
    the tree. Each entry of this difference dictionary is a tuple (val1, val2) when val1 != val2.

    """
    diff = {}
    for key in dict1.keys():
        if key not in dict2.keys():
            diff[key] = (dict1[key], None)
        else:
            if isinstance(dict1[key], dict):
                inner_diff = compare_dictionaries(dict1[key], dict2[key])
                if len(inner_diff) > 0:
                    diff[key] = inner_diff
            elif isinstance(dict1[key], np.ndarray):
                if not np.allclose(dict1[key] - dict2[key], 0, rtol=1e-6, atol=1e-6):
                    diff[key] = (dict1[key], dict2[key])
            else:
                if dict1[key] != dict2[key]:
                    diff[key] = (dict1[key], dict2[key])

    for key in dict2.keys():
        if key not in dict1.keys():
            diff[key] = (None, dict2[key])

    return diff


class TemporaryNumpySeed:
    def __init__(self, seed):
        self.seed = seed
        self.state = None

    def __enter__(self):
        self.state = np.random.get_state()
        np.random.seed(self.seed)

    def __exit__(self, *args):
        np.random.set_state(self.state)


class ConfigException(Exception):
    """Exception raised for errors in the configuration."""

    def __init__(self, message):
        super().__init__(message)



import functools
from collections import OrderedDict

import numpy as np
from escnn.group import Representation, directsum

from morpho_symm.utils.algebra_utils import permutation_matrix


def decom_signal_into_isotypic_components(signal: np.ndarray, rep: Representation):
    """Decompose a signal into its isotypic components.

    This function takes a signal and a group representation and returns the signal decomposed into its isotypic
    components, in the isotypic basis and the original basis.

    Args:
        signal: (... , rep.size) array of signals to be decomposed.
        rep: (Representation) Representation of the vector space in which the signal lives in.

    Returns:
        iso_comp_signals: (OrderedDict) Dictionary of isotypic components of the signal in the isotypic basis.
        iso_comp_signals_orig_basis: (OrderedDict) Dictionary of isotypic components of the signal in the original
        basis.
    """
    rep_iso = isotypic_decomp_representation(rep)
    Q_iso2orig = rep_iso.change_of_basis  # Change of basis from isotypic basis to original basis
    Q_orig2iso = rep_iso.change_of_basis_inv  # Change of basis from original basis to isotypic basis
    assert signal.shape[-1] == rep.size, f"Expected signal shape to be (..., {rep.size}) got {signal.shape}"

    signal_iso = np.einsum("...ij,...j->...i", Q_orig2iso, signal)

    isotypic_representations = rep_iso.attributes["isotypic_reps"]

    # Compute the dimensions of each isotypic subspace
    cum_dim = 0
    iso_comp_dims = {}
    for irrep_id, iso_rep in isotypic_representations.items():
        iso_space_dims = range(cum_dim, cum_dim + iso_rep.size)
        iso_comp_dims[irrep_id] = iso_space_dims
        cum_dim += iso_rep.size

    # Separate the signal into isotypic components, by masking the signal outside of each isotypic subspace
    iso_comp_signals = OrderedDict()
    for irrep_id, _ in isotypic_representations.items():
        iso_dims = iso_comp_dims[irrep_id]
        iso_comp_signals[irrep_id] = signal_iso[..., iso_dims]

    iso_comp_signals_orig_basis = OrderedDict()
    # Compute the signals of each isotypic component in the original basis
    for irrep_id, _ in isotypic_representations.items():
        iso_dims = iso_comp_dims[irrep_id]
        Q_isocomp2orig = Q_iso2orig[:, iso_dims]  # Change of basis from isotypic component basis to original basis
        iso_comp_signals_orig_basis[irrep_id] = np.einsum(
            "...ij,...j->...i", Q_isocomp2orig, iso_comp_signals[irrep_id]
        )

    # Check that the sum of the isotypic components is equal to the original signal
    rec_signal = np.sum([iso_comp_signals_orig_basis[irrep_id] for irrep_id in isotypic_representations.keys()], axis=0)
    assert np.allclose(rec_signal, signal), (
        f"Reconstructed signal is not equal to original signal. Error: {np.linalg.norm(rec_signal - signal)}"
    )

    return iso_comp_signals, iso_comp_signals_orig_basis


def isotypic_decomp_representation(rep: Representation) -> Representation:
    """Returns a representation in a "symmetry enabled basis" (a.k.a Isotypic Basis).

    Takes a representation with an arbitrary basis (i.e., arbitrary change of basis and an arbitrary order of
    irreducible representations in the escnn Representation) and returns a new representation in which the basis
    is changed to a "symmetry enabled basis" (a.k.a Isotypic Basis). That is a representation in which the
    vector space is decomposed into a direct sum of Isotypic Subspaces. Each Isotypic Subspace is a subspace of the
    original vector space with a subspace representation composed of multiplicities of a single irreducible
    representation. In oder words, each Isotypic Subspace is a subspace with a subgroup of symmetries of the original
    vector space's symmetry group.

    Args:
        rep (Representation): Input representation in any arbitrary basis.

    Returns: A `Representation` with a change of basis exposing an Isotypic Basis (a.k.a symmetry enabled basis).
        The instance of the representation contains an additional attribute `isotypic_subspaces` which is an
        `OrderedDict` of representations per each isotypic subspace. The keys are the active irreps' ids associated
        with each Isotypic subspace.
    """
    symm_group = rep.group
    potential_irreps = rep.group.irreps()
    isotypic_subspaces_indices = {irrep.id: [] for irrep in potential_irreps}

    for pot_irrep in potential_irreps:
        cur_dim = 0
        for rep_irrep_id in rep.irreps:
            rep_irrep = symm_group.irrep(*rep_irrep_id)
            if rep_irrep == pot_irrep:
                isotypic_subspaces_indices[rep_irrep_id].append(list(range(cur_dim, cur_dim + rep_irrep.size)))
            cur_dim += rep_irrep.size

    # Remove inactive Isotypic Spaces
    for irrep in potential_irreps:
        if len(isotypic_subspaces_indices[irrep.id]) == 0:
            del isotypic_subspaces_indices[irrep.id]

    # Each Isotypic Space will be indexed by the irrep it is associated with.
    active_isotypic_reps = {}
    for irrep_id, indices in isotypic_subspaces_indices.items():
        irrep = symm_group.irrep(*irrep_id)
        multiplicities = len(indices)
        active_isotypic_reps[irrep_id] = Representation(
            group=rep.group,
            irreps=[irrep_id] * multiplicities,
            name=f"IsoSubspace {irrep_id}",
            change_of_basis=np.identity(irrep.size * multiplicities),
            supported_nonlinearities=irrep.supported_nonlinearities,
        )

    # Impose canonical order on the Isotypic Subspaces.
    # If the trivial representation is active it will be the first Isotypic Subspace.
    # Then sort by dimension of the space from smallest to largest.
    ordered_isotypic_reps = OrderedDict(sorted(active_isotypic_reps.items(), key=lambda item: item[1].size))
    if symm_group.trivial_representation.id in ordered_isotypic_reps.keys():
        ordered_isotypic_reps.move_to_end(symm_group.trivial_representation.id, last=False)

    # Required permutation to change the order of the irreps. So we obtain irreps of the same type consecutively.
    oneline_permutation = []
    for irrep_id, iso_rep in ordered_isotypic_reps.items():
        idx = isotypic_subspaces_indices[irrep_id]
        oneline_permutation.extend(idx)
    oneline_permutation = np.concatenate(oneline_permutation)
    P_in2iso = permutation_matrix(oneline_permutation)

    Q_iso = rep.change_of_basis @ P_in2iso.T
    rep_iso_basis = directsum(list(ordered_isotypic_reps.values()), name=rep.name + "-Iso", change_of_basis=Q_iso)

    iso_supported_nonlinearities = [iso_rep.supported_nonlinearities for iso_rep in ordered_isotypic_reps.values()]
    rep_iso_basis.supported_nonlinearities = functools.reduce(set.intersection, iso_supported_nonlinearities)
    rep_iso_basis.attributes["isotypic_reps"] = ordered_isotypic_reps

    return rep_iso_basis


def isotypic_basis(representation: Representation, multiplicity: int = 1, prefix=""):
    rep_iso = isotypic_decomp_representation(representation)

    iso_reps = OrderedDict()
    iso_range = OrderedDict()

    start_dim = 0
    for iso_irrep_id, reg_rep_iso in rep_iso.attributes["isotypic_reps"].items():
        iso_reps[iso_irrep_id] = directsum([reg_rep_iso] * multiplicity, name=f"{prefix}_IsoSpace{iso_irrep_id}")
        iso_range[iso_irrep_id] = range(start_dim, start_dim + iso_reps[iso_irrep_id].size)
        start_dim += iso_reps[iso_irrep_id].size

    assert rep_iso.size * multiplicity == sum([iso_rep.size for iso_rep in iso_reps.values()])

    return iso_reps, iso_range  # Dict[key:id_space -> value: rep_iso_space]



import os
import pathlib
import time

import hydra
import numpy as np
from omegaconf import DictConfig
from scipy.spatial.transform import Rotation
from utils.pybullet_visual_utils import (
    change_robot_appearance,
    display_robots_and_vectors,
    get_mock_ground_reaction_forces,
    render_orbiting_animation,
)
from utils.robot_utils import load_symmetric_system

import morpho_symm
from morpho_symm.utils.pybullet_visual_utils import configure_bullet_simulation

os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"


@hydra.main(config_path="cfg/", config_name="config_visualization", version_base="1.3")
def main(cfg: DictConfig):
    """Visualize the effect of DMSs transformations in 3D animation.

    This script visualizes the DMSs transformations on robot state and on proprioceptive and exteroceptive measurements.
    """
    cfg.robot.seed = cfg.robot.seed if cfg.robot.seed >= 0 else np.random.randint(0, 1000)
    np.random.seed(cfg.robot.seed)
    # Get robot instance, along with representations of the symmetry group on the Euclidean space (in which the robot
    # base B evolves in) and Joint Space (in which the internal configuration of the robot evolves in).
    robot, G = load_symmetric_system(robot_cfg=cfg.robot, debug=cfg.debug_joints)

    # Get the group representations of joint-state space, and Euclidean space.
    rep_QJ = G.representations["Q_js"]  # rep_QJ(g) is a permutation matrix ∈ R^nqj
    rep_TqQJ = G.representations["TqQ_js"]  # rep_TqQJ(g) is a permutation matrix ∈ R^nvj
    rep_Ed = G.representations["E3"]  # rep_Ed(g) is a homogenous transformation matrix ∈ R^(3+1)x(3+1)
    rep_Rd = G.representations["R3"]  # rep_Rd(g) is an orthogonal matrix ∈ R^3x3
    rep_Rd_pseudo = G.representations["R3_pseudo"]  # rep_Rd_pseudo(g) is an orthogonal matrix ∈ R^3x3
    rep_euler_xyz = G.representations["euler_xyz"]  # rep_euler_xyz(g) is an euler angle vector ∈ R^3

    # Configuration of the 3D visualization -------------------------------------------------------------------------
    # Not really relevant to understand.
    offset = max(0.2, 1.8 * robot.hip_height)  # Offset between robot base and reflection planes.
    pb = configure_bullet_simulation(gui=cfg.gui, debug=cfg.debug_joints)  # Load pybullet environment
    robot.configure_bullet_simulation(pb, world=None)  # Load robot in pybullet environment
    change_robot_appearance(pb, robot, change_color=cfg.robot.tint_bodies)  # Add color and style to boring grey robots

    # Get initial random configuration of the system
    q, v = robot.get_init_config(random=True, angle_sweep=cfg.robot.angle_sweep, fix_base=cfg.robot.fix_base)
    # q[7:] *= 0
    rB0 = np.array([-offset if G.order() != 2 else 0, -offset] + [robot.hip_height * 1.5])
    q[:3] = rB0  # Place base of the robot with some `offset` from origin.
    robot.reset_state(q, v)  # Reset robot state in pybullet and pinocchio.
    # --------------------------------------------------------------------------------------------------------------

    # NOTATION:
    # We use the following code notation which tries to replicate as much as possible the notation of the paper:
    # [Variable][Frame]_[reference frame of variable] e.g, RB_w : Rotation (R) of Base (B) in World (w) coordinates.
    # For compactness the reference frame is omitted if variables are measured w.r.t (w) world frame.
    # - `Ed`: Euclidean space of d dimensions
    # - `QJ`: Joint Space of the robot
    # - `TQJ`: Tangent space to the joint space.
    # - `r`: Vector in `Ed` describing the position of a frame of the robot bodies. e.g. rB: Robot base position
    # - `R`: SO3 matrix describing a rotation in E3. e.g. RB: Robot base orientation
    # - `X`: Homogenous Matrix 4x4 transformation holding a rotation matrix and a translation X = [[R , r], [0, 1]]

    # Get the robot's base configuration XB ∈ Ed as a homogenous transformation matrix.
    XB = robot.get_base_configuration()
    base_ori_euler_xyz = Rotation.from_matrix(XB[:3, :3]).as_euler("xyz", degrees=True)

    # Get joint space position and velocity coordinates  (q_js, v_js) | q_js ∈ QJ, dq_js ∈ TqQJ
    q_js, v_js = robot.get_joint_space_state()

    # To visualize how the symmetries of the robotic system affect proprioceptive and exteroceptive measurements we use:
    # - Proprioceptive: `hg_B` The center of mass linear and angular momentum of the robot, and `f1/2` contact forces
    # - Exteroceptive: We will assume to have measurements of the terrain elevation and orientation.
    # Consider the robot Center of Mass (CoM) momentum in base `B` coordinates. A vector and a pseudo-vector.
    hg_B = np.array([0.1, 0.1, 0.0, -0.0, -0.2, -0.2])  # Set to fix value for visualization.
    # The representation for the linear `l` and angular `k` components  of Center of Mass (CoM) momentum `h=[l,k]` is:
    rep_h = rep_Rd + rep_Rd_pseudo  # Additions of representations amounts to block-diagonal matrix concatenation.

    # Define mock surface orientations `Rf_w`, force contact points `rf_w` and contact forces `f_w`
    Rf1, Rf2, f1, f2, rf1, rf2 = get_mock_ground_reaction_forces(pb, robot, cfg.robot)

    # Main part of the script. =======================================================================================
    # Start by defining the dict representing the orbits of all the proprioceptive and exteroceptive measurements
    # An orbit is all unique symmetric states of a variable G·z = {g·z | g in G}
    e = G.identity  # Identity element of the group
    orbit_q_js, orbit_v_js = {e: q_js}, {e: v_js}
    orbit_hg_B, orbit_XB_w = {e: hg_B}, {e: XB}
    orbit_f1, orbit_f2, orbit_rf1, orbit_rf2 = (
        {e: f1},
        {e: f2},
        {e: rf1},
        {e: rf2},
    )
    orbit_Rf1, orbit_Rf2 = {e: Rf1}, {e: Rf2}
    orbit_ori_euler_xyz = {e: base_ori_euler_xyz}
    # For each symmetry action g ∈ G, we get the representations of the action in the relevant vector spaces to
    # compute the symmetric states of robot configuration and measurements.
    for g in G.elements[1:]:
        # Get symmetric joint-space state (g.q_js, g.v_js), and CoM momentum g·h=[g·l, g·k] ---------------------------
        # gx_w, gh_B = (rep_x(g) @ x).astype(x.dtype), (rep_h(g) @ hg_B).astype(hg_B.dtype)
        orbit_q_js[g], orbit_v_js[g] = rep_QJ(g) @ q_js, rep_TqQJ(g) @ v_js
        orbit_hg_B[g] = rep_h(g) @ hg_B

        # Compute new robot base configuration -----------------------------------------------------------------------
        # gXB_w = rep_Ed(g) @ RB_w @ np.linalg.inv(rep_Ed(g))
        gXB = rep_Ed(g) @ XB @ rep_Ed(g).T
        orbit_XB_w[g] = gXB  # Add new robot base configuration (homogenous matrix) to the orbit of base configs.
        orbit_ori_euler_xyz[g] = Rotation.from_matrix(gXB[:3, :3]).as_euler("xyz", degrees=True)
        # If people use euler xyz angles to represent the orientation of the robot base, we can also compute the
        # symmetric states of the robot base orientation:
        g_euler_xyz = rep_euler_xyz(g) @ base_ori_euler_xyz
        # Check the analytic transformation to elements of SO(3) is equivalent to the transformation in euler xyz angles
        g_euler_xyz_true = Rotation.from_matrix(gXB[:3, :3]).as_euler("xyz", degrees=True)
        assert np.allclose(g_euler_xyz, g_euler_xyz_true, rtol=1e-6, atol=1e-6)

        # Use symmetry representations to get symmetric versions of Euclidean vectors, representing measurements of data
        orbit_f1[g], orbit_f2[g] = rep_Rd(g) @ f1, rep_Rd(g) @ f2
        orbit_rf1[g] = (rep_Ed(g) @ np.concatenate((rf1, np.ones(1))))[:3]  # using homogenous coordinates
        orbit_rf2[g] = (rep_Ed(g) @ np.concatenate((rf2, np.ones(1))))[:3]

        # Apply transformations to the terrain elevation estimations/measurements
        orbit_Rf1[g] = rep_Rd(g) @ Rf1 @ rep_Rd(g).T
        orbit_Rf2[g] = rep_Rd(g) @ Rf2 @ rep_Rd(g).T

    for g in G.elements:
        print(f"Element: {g} euler_xyz(g): \n{rep_euler_xyz(g)}")
        print(f"Element: {g} Rd_pseudo(g): \n{rep_Rd_pseudo(g)}")

    # Visualization of orbits of robot states and of data ==========================================================
    # Use Ctrl and mouse-click+drag to rotate the 3D environment.
    display_robots_and_vectors(
        pb,
        robot,
        group=G,
        base_confs=orbit_XB_w,
        orbit_q_js=orbit_q_js,
        orbit_v_js=orbit_v_js,
        orbit_com_momentum=orbit_hg_B,
        forces=[orbit_f1, orbit_f2],
        forces_points=[orbit_rf1, orbit_rf2],
        surface_normals=[orbit_Rf1, orbit_Rf2],
        tint=cfg.robot.tint_bodies,
        draw_floor=cfg.robot.draw_floor,
    )

    root_path = pathlib.Path(morpho_symm.__file__).parents[1].absolute()
    if cfg.make_gif:
        # To get the optimal visualization you might need to play a bit with the rendering parameters for each robot.
        # this params seem to work mostly ok for all.
        cam_distance = offset * 5
        cam_target_pose = [0, 0, 0]
        save_path = root_path / "docs/static/animations"
        save_path.mkdir(exist_ok=True)
        render_orbiting_animation(
            pb,
            cam_target_pose=cam_target_pose,
            cam_distance=cam_distance,
            save_path=save_path,
            anim_time=10,
            fps=15,
            periods=1,
            init_roll_pitch_yaw=(0, 35, 0),
            invert_roll="dh" in cfg.robot.group_label.lower(),
            pitch_sin_amplitude=20,
            file_name=f"{robot.name}-{G.name}-symmetries_anim_static",
            gen_gif=True,
            gen_imgs=False,
        )
        print("Done enjoy your gif :). I hope you learned something new")
    elif cfg.make_imgs:
        cam_distance = offset * 6
        cam_target_pose = [0, 0, 0]
        save_path = root_path / f"paper/images/{cfg.robot.name}"
        save_path.mkdir(exist_ok=True)
        render_orbiting_animation(
            pb,
            cam_target_pose=cam_target_pose,
            cam_distance=cam_distance,
            anim_time=2,
            fps=2,
            periods=1,
            pitch_sin_amplitude=0,
            init_roll_pitch_yaw=(0, 90, 0) if G.order() > 2 else (0, 0, 0),
            invert_roll="dh" in cfg.robot.group_label.lower(),
            save_path=save_path,
            gen_gif=False,
            gen_imgs=True,
        )
    if cfg.gui:
        while True:
            time.sleep(0.1)

        pb.disconnect()


if __name__ == "__main__":
    main()



import time

import numpy as np
from omegaconf import OmegaConf
from pynput import keyboard
from scipy.spatial.transform import Rotation
from utils.pybullet_visual_utils import change_robot_appearance, spawn_robot_instances
from utils.robot_utils import load_symmetric_system

from morpho_symm.robots.PinBulletWrapper import PinBulletWrapper
from morpho_symm.utils.abstract_harmonics_analysis import isotypic_decomp_representation
from morpho_symm.utils.pybullet_visual_utils import configure_bullet_simulation


def on_key_press(key):
    try:
        # Get the pressed key and add it to the global list of new characters
        if key.char.isdigit():
            num_pressed.append(int(key.char))
        else:
            new_command.append(key.char)
    except AttributeError:
        # Ignore special keys without a char attribute
        pass


# Set up the listener and the list to store new characters
new_command = []
num_pressed = []
listener = keyboard.Listener(on_press=on_key_press)
listener.start()


def generate_dof_motions(robot: PinBulletWrapper, angle_sweep=0.5):
    q0, _ = robot.get_init_config(random=False)
    phases = np.random.uniform(0, 2 * np.pi, robot.nq)[..., None]
    amplitudes = np.random.uniform(0, angle_sweep, robot.nq)[..., None]
    max_period, min_period = 6, 3
    periods = np.random.randint(min_period, max_period, robot.nq)[..., None]
    q_period = np.lcm.reduce(periods).item()
    t = np.linspace(0, q_period, q_period * 10)[None, ...]

    s_t = (np.sin(2 * np.pi * (1 / periods) * t + phases) * amplitudes).T
    q = np.zeros((t.size, robot.nq - 1))
    q[:, :3] = s_t[..., :3] * 0.2 / s_t[..., :3].max()
    q[:, 3:6] = s_t[..., 3:6] * (30 * np.pi / 180) / s_t[..., 3:6].max()
    q[:, 7:] = s_t[..., 7:-1]
    return q[:, : robot.nq]  # We return orientation in euler angles


def coord_max2min(q):
    pos = q[:3]
    ori = Rotation.from_quat(q[3:7]).as_euler("xyz")
    return np.concatenate([pos, ori, q[7:]])


def coord_min2max(q):
    pos = q[:3]
    ori = Rotation.from_euler("xyz", q[3:6]).as_quat()
    return np.concatenate([pos, ori, q[6:]])


if __name__ == "__main__":
    """
    This example script shows how to apply the isotypic decomposition of a floating-base robot robot's configuration's
    space Q. The space is assumed to be composed of a MINIMAL set of generalized coordinates, defining a one-to-one map
    to the robot's degrees of freedom. Therefore we assume Q = [base_pos, base_ori, q_js] where:
        - base_pos: The base position (3,)
        - base_ori: The base orientation (3,) in Euler angles xyz (roll, pitch, yaw)
        - q_js: The joint space positions (n_dof,)

    Note: To select a different robot pass the robot name as an argument to the script.
    ```
    python robot_harmonic_decomposition_mini_cheetah.py robot=mini_cheetah  # solo, solo-4, atlas, baxter, anymal_c
    ```
    
    Note: Pinocchio defines the position configuration generalized coordinates in maximal/redundant coordinates.
    For instance the base orientation is defined in quaternions and positions of revolute unbounded joints are defined
    as points in the unit circle theta -> [cos(theta), sin(theta)]. For the isotypic decomposition we are required to
    work on minimal coordinates. So be mindful that the representation defined above is not the one acting on
    pinocchio's `q` vector.
    
    """
    # Get the command line arguments, the only expected is robot=<robot_name>
    conf = OmegaConf.from_cli()
    robot_name = conf.get("robot", "atlas")  # Default robot is mini_cheetah

    # Load the robot instance its symmetry group G ____________________________________________________________________
    robot, G = load_symmetric_system(robot_name=robot_name)
    # Now we obtain the group representations required to define the representation of the configuration space Q.
    rep_R3 = G.representations["R3"]
    rep_euler_xyz = G.representations["euler_xyz"]
    rep_Q_js = G.representations["TqQ_js"]

    # Define the group representation of the configuration space Q. ___________________________________________________
    # rep_Q := T @ (rep_Q_iso_1 + ..., rep_Q_iso_k) @ T_inv,                 T: Q -> Q
    rep_Q = isotypic_decomp_representation(rep_R3 + rep_euler_xyz + rep_Q_js)
    # Get the matrix change of basis mapping between the isotypic basis and the canonical basis of Q.
    T = rep_Q.change_of_basis  # T: Q -> Q  from isotypic basis to canonical basis
    T_inv = rep_Q.change_of_basis_inv  # T_inv: Q -> Q from canonical basis to isotypic basis
    num_iso_components = len(rep_Q.attributes["isotypic_reps"])

    # Define the orthogonal projectors of the state position to each isotypic subspace. _______________________________
    # The process of projecting the state into its isotypic components is done by applying a change to the isotypic
    # basis, isolating the values associated to the isotypic subspace and then returning to the canonical basis.
    # P_i = T @ diag(mask_iso_i) @ T_inv, such that q^(iso_i) = P_1 @ q
    iso_projectors = {}
    tmp_dim = 0
    for iso_id, iso_rep in rep_Q.attributes["isotypic_reps"].items():
        mask = np.zeros(rep_Q.size)
        mask[tmp_dim : tmp_dim + iso_rep.size] = 1
        iso_projectors[iso_id] = T @ np.diag(mask) @ T_inv
        print(f"Isotypic component {iso_id} has dimension {iso_rep.size}")
        tmp_dim += iso_rep.size
    assert rep_Q.size == robot.nv, f"Conf space rep dimension {rep_Q.size} != |Q|:={robot.nv}"

    # Illustrate how the mass matrix at any random configuration q can be block-diagonalized by change of basis T_inv.
    q0, dq0 = robot.get_init_config(random=True, angle_sweep=20 * np.pi / 180)

    # Ensure the robot's Joint-Space mass matrix is a G-equivariant matrix-valued function M(g * q) = g * M(q) * g^T
    for g in G.elements:
        if g == G.identity:
            continue
        Mq0 = robot.pinocchio_robot.mass(q0)
        g_q0 = coord_min2max(rep_Q(g) @ coord_max2min(q0))
        Mg_q0 = robot.pinocchio_robot.mass(g_q0)
        Mg_g0_pred = rep_Q(g) @ Mq0 @ rep_Q(~g)
        Mg_q0_iso = T @ Mg_q0 @ T_inv
        rep_Q_g_iso = T_inv @ rep_Q(g) @ T
        # Ignore the world position coordinates.
        assert np.allclose(Mg_q0[:3, :3], Mg_g0_pred[:3, :3], atol=1e-6, rtol=1e-6)

    # Configure visualization animation _______________________________________________________________________________
    pb = configure_bullet_simulation(gui=True)
    robot.configure_bullet_simulation(pb)
    change_robot_appearance(pb, robot)

    base_pos, base_ori_quat_xyzw = q0[:3], q0[3:7]
    robot.reset_state(q0, dq0)

    # For each isotypic component we spawn a robot instance in order to visualize the effect of the decomposition
    base_positions = np.asarray([base_pos] * num_iso_components)
    base_positions[:, 0] = -4.0 * robot.hip_height
    base_positions[:, 1] = np.linspace(0, 2.8 * robot.hip_height * num_iso_components, num_iso_components)
    base_positions[:, 1] -= np.max(base_positions[:, 1]) / 2
    # base_positions[:, 2] += 0.5

    iso_robots = spawn_robot_instances(
        robot,
        bullet_client=pb,
        base_positions=base_positions,
        base_orientations=[base_ori_quat_xyzw] * num_iso_components,
        tint=True,
        alpha=0.5,
    )
    # _________________________________________________________________________________________________________________

    # Define the state trajectories that will be used to visualize the isotypic decomposition. ________________________
    traj_q = generate_dof_motions(robot, angle_sweep=90 * np.pi / 180)
    time_horizon = traj_q.shape[0]
    # Project the trajectories into the isotypic components.
    traj_q_iso = {}
    for iso_id, iso_rep in rep_Q.attributes["isotypic_reps"].items():
        traj_q_iso[iso_id] = np.einsum("ij, tj -> ti", iso_projectors[iso_id], traj_q)
    # Check that q(t) := sum_i^n_iso q^(iso_i)(t)
    rec_traj_q_iso = np.sum([traj_q_iso[iso_id] for iso_id in rep_Q.attributes["isotypic_reps"].keys()], axis=0)
    assert np.allclose(rec_traj_q_iso, traj_q, atol=1e-6, rtol=1e-6), (
        f"Reconstruction error {np.linalg.norm(rec_traj_q_iso - traj_q)}"
    )

    # Visualize the robot state trajectory and its projection to each isotypic subspace _______________________________
    t_idx = 0
    fps = 30
    timescale = 0.5
    time_shift = 1
    frames = []
    while True:
        # Visualize the robot configuration
        traj_q_js = traj_q[t_idx, 6:]
        base_ori = Rotation.from_euler("xyz", traj_q[t_idx, 3:6]).as_quat()
        robot.reset_state(q=np.concatenate([base_pos + traj_q[t_idx, :3], base_ori, traj_q_js]), v=np.zeros(robot.nv))

        # Visualize the projections of the position configuration into the isotypic components.
        for n, iso_robot, iso_id in zip(range(len(iso_robots)), iso_robots, traj_q_iso.keys()):
            traj_q_js_iso_i = traj_q_iso[iso_id][t_idx, 6:]
            base_ori_iso_i = Rotation.from_euler("xyz", traj_q_iso[iso_id][t_idx, 3:6]).as_quat()
            iso_robot.reset_state(
                q=np.concatenate([base_positions[n] + traj_q_iso[iso_id][t_idx, :3], base_ori_iso_i, traj_q_js_iso_i]),
                v=np.zeros(robot.nv),
            )

        # Process new keyboard commands.
        if new_command:
            keys = new_command.copy()
            new_command.clear()
            if keys == ["t"]:  # Pauses/Resumes the replay
                time_shift = 1 if time_shift == 0 else 0

        # Update the time index
        t_idx += time_shift
        if t_idx == time_horizon:
            t_idx = 0
        time.sleep(1 / fps / timescale)
    pb.disconnect()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 11/3/22
# @Author  : Daniel Ordonez
# @email   : daniels.ordonez@gmail.com
from __future__ import annotations

import logging
import re
import warnings
from pathlib import Path
from typing import Optional

import escnn
import escnn.group
import numpy as np
from escnn.group import CyclicGroup, DihedralGroup, DirectProductGroup, Group, Representation, change_basis
from omegaconf import DictConfig

import morpho_symm
from morpho_symm.utils.algebra_utils import gen_permutation_matrix, permutation_matrix
from morpho_symm.utils.mysc import ConfigException, load_config_hierarchy
from morpho_symm.utils.pybullet_visual_utils import (
    change_robot_appearance,
    configure_bullet_simulation,
    listen_update_robot_sliders,
    setup_debug_sliders,
)
from morpho_symm.utils.rep_theory_utils import escnn_representation_form_mapping, group_rep_from_gens

log = logging.getLogger("MorphoSymm")


def get_escnn_group(group_label: str):
    """Get the ESCNN group object from the group label in the config file."""
    label_pattern = r"([A-Za-z]+)(\d+)"
    assert group_label is not None, "Group label unspecified"
    match = re.match(label_pattern, group_label)
    if match:
        group_class = match.group(1)
        order = int(match.group(2))
    else:
        raise AttributeError(f"Group label {group_label} is not a known group label (Dn: Dihedral, Cn: Cyclic) order n")

    group_axis = np.array([0, 0, 1])
    subgroup_id = np.zeros_like(group_axis, dtype=bool).astype(object)
    if group_class.lower() == "d":  # Dihedral
        # Define the symmetry space using presets from ESCNN
        # subgroup_id[group_axis == 1] = order
        symmetry_space = escnn.gspaces.dihedralOnR3(n=order // 2, axis=0.0)
    elif group_class.lower() == "c":  # Cyclic
        assert order >= 2, f"Order of cyclic group must be greater than 2, got {order}"
        subgroup_id[group_axis == 1] = order
        symmetry_space = escnn.gspaces.GSpace3D(tuple(subgroup_id))
    elif group_class.lower() == "k":  # Klein four group
        is_planar_subgroup = True
        symmetry_space = escnn.gspaces.GSpace3D(sg_id=(is_planar_subgroup, False, 2))
    elif group_class.lower() == "dh":  # Dn x C2. Group of regular rectangular cuboid
        symmetry_space = escnn.gspaces.GSpace3D(sg_id=(True, True, order))
    else:
        raise NotImplementedError(f"We have to implement the case of {group_label}")

    return symmetry_space


def load_symmetric_system(
    robot_cfg: Optional[DictConfig] = None,
    robot_name: Optional[str] = None,
    debug=False,
    return_robot=True,
    joint_space_order: list[str] = None,
) -> [escnn.group.Group]:
    """Utility function to get the symmetry group and representations of a robotic system defined in config.

    This function loads the robot into pinocchio, and generate the symmetry group representations for the following
    spaces:
        1. The joint-space (Q_js), known as the space of generalized position coordinates.
        2. The joint-space tangent space (TqQ_js), known as the space of generalized velocities.

    Args:
        robot_name: (str): (Optional) name of the robot configuration file in `cfg/robot/` to load.
        robot_cfg (DictConfig): (Optional) configuration parameters of the robot. Check `cfg/robot/`
        debug (bool): if true we load the robot into an interactive simulation session to visually inspect URDF
        return_robot (bool): if true we return the robot instance loaded in pyBullet.
        joint_space_order (list[str]): (Optional) Order of the joint-space coordinates. Must be a list of joint names matching those of the urdf file.

    Returns:
        robot (PinBulletWrapper): instance with the robot loaded in pinocchio and ready to be loaded in pyBullet
        G (escnn.group.Group): Instance of the symmetry Group of the robot. The representations for Q_js, TqQ_js and
        Ed are
        added to the list of representations of the group.
    """
    assert robot_cfg is not None or robot_name is not None, (
        "Either a robot configuration file or a robot name must be provided."
    )
    if robot_cfg is None:
        # Load hydra configs without hydra.
        path_cfg = Path(morpho_symm.__file__).parent / "cfg" / "robot"
        robot_cfg = load_config_hierarchy(cfg_path=path_cfg / f"{robot_name}.yaml")

    robot_name = str.lower(robot_cfg.name)

    symmetry_space = get_escnn_group(group_label=robot_cfg.group_label)
    G = symmetry_space.fibergroup

    # Select the field for the representations.
    rep_field = float if robot_cfg.rep_fields.lower() != "complex" else complex

    # Assert the required representations are present in the robot configuration.
    if "permutation_Q_js" not in robot_cfg:
        raise ConfigException(
            f"Configuration file for {robot_name} must define the field `permutation_Q_js`, "
            f"describing the joint space permutation per each non-trivial group's generator."
        )
    if "permutation_TqQ_js" not in robot_cfg:
        raise ConfigException(
            f"Configuration file for {robot_name} must define the field `permutation_TqQ_js`, "
            f"describing the tangent joint-space permutation per each non-trivial group's generator."
        )

    reps_in_cfg = [k.split("permutation_")[1] for k in robot_cfg if "permutation" in k]

    for rep_name in reps_in_cfg:
        try:
            perm_list = list(robot_cfg[f"permutation_{rep_name}"])
            rep_dim = len(perm_list[0])
            reflex_list = list(robot_cfg[f"reflection_{rep_name}"])
            assert len(perm_list) == len(reflex_list), (
                f"Found different number of permutations and reflections for {rep_name}"
            )
            assert len(perm_list) >= len(G.generators), (
                f"Found {len(perm_list)} element reps for {rep_name}, Expected {len(G.generators)} generators for {G}"
            )
            # Generate ESCNN representation of generators
            gen_rep = {}
            for h, perm, refx in zip(G.generators, perm_list, reflex_list):
                assert len(perm) == len(refx) == rep_dim, f"For {h}: len(perm)={len(perm)}, len(refx)={len(refx)}"
                refx = np.array(refx, dtype=rep_field)
                gen_rep[h] = gen_permutation_matrix(oneline_notation=perm, reflections=refx)
            # Generate the entire group
            rep = group_rep_from_gens(G, rep_H=gen_rep)
            rep.name = rep_name
            G.representations.update({rep_name: rep})
        except Exception as e:
            raise ConfigException(f"Error in the definition of the representation {rep_name}") from e

    rep_Q_js = G.representations["Q_js"]
    rep_TqQ_js = G.representations.get("TqQ_js", None)
    rep_TqQ_js = rep_Q_js if rep_TqQ_js is None else rep_TqQ_js

    # Create the representation of isometries on the Euclidean Space in d dimensions.
    # This adds `O3` and `E3` representations to the group
    robot_rep_names = ["Q_js", "TqQ_js", "E3", "R3", "R3_pseudo", "E3_pseudo"]
    for rep_name in robot_rep_names:
        if rep_name in G.representations:
            G.representations.pop(rep_name)
            warnings.warn(
                f"Representation {rep_name} found while loading the robot {robot_name}. "
                f"This happens when you load multiple robots in the same session. "
                f"The old representation will be overwritten."
            )

    rep_R3, rep_E3, rep_R3pseudo, rep_E3pseudo = generate_euclidean_space_representations(G)

    # Define the representation of the rotation matrix R that transforms the base orientation.
    rep_rot_flat = {}
    for h in G.elements:
        rep_rot_flat[h] = np.kron(rep_R3(h), rep_R3(~h).T)
    rep_rot_flat = escnn_representation_form_mapping(G, rep_rot_flat)
    rep_rot_flat.name = "SO3_flat"

    log.info(f"Loaded robot {robot_name}, with defined group representations:")
    for name, rep in G.representations.items():
        log.info(f"\t {name}: dimension: {rep.size}")

    if return_robot or joint_space_order is not None:
        from morpho_symm.robots.PinBulletWrapper import PinBulletWrapper

        # We allow symbolic expressions (e.g. `np.pi/2`) in the `q_zero` and `init_q`.
        q_zero = (
            np.array([eval(str(s)) for s in robot_cfg.q_zero], dtype=float) if robot_cfg.q_zero is not None else None
        )
        init_q = (
            np.array([eval(str(s)) for s in robot_cfg.init_q], dtype=float) if robot_cfg.init_q is not None else None
        )
        robot = PinBulletWrapper(
            robot_name=robot_name,
            init_q=init_q,
            q_zero=q_zero,
            hip_height=robot_cfg.hip_height,
            endeff_names=robot_cfg.endeff_names,
            fixed_base=robot_cfg.fix_base,
        )

        default_joint_order = robot.joint_space_names
        if joint_space_order is not None and default_joint_order != joint_space_order:
            assert len(joint_space_order) == len(default_joint_order), (
                f"|joint_space_order|={len(joint_space_order)} != |default_joint_order|={len(default_joint_order)}"
            )
            assert set(joint_space_order) == set(default_joint_order), (
                f"Invalid joint order, missing joints:  {default_joint_order - joint_space_order}"
            )
            log.info(f"Changing joint-space order to match the required order {joint_space_order}")
            # Oneline notation of the permutation from default to required joint order
            perm = np.array([default_joint_order.index(j) for j in joint_space_order])
            P = permutation_matrix(oneline_notation=perm)

            rep_Q_js = change_basis(
                rep_Q_js, P, name="Q_js", supported_nonlinearities=rep_Q_js.supported_nonlinearities
            )
            rep_TqQ_js = change_basis(
                rep_TqQ_js, P, name="TqQ_js", supported_nonlinearities=rep_TqQ_js.supported_nonlinearities
            )

        if debug:
            pb = configure_bullet_simulation(gui=True, debug=debug)
            robot.configure_bullet_simulation(pb, world=None)
            change_robot_appearance(pb, robot)
            setup_debug_sliders(pb, robot)
            listen_update_robot_sliders(pb, robot)

        dimQ_js, dimTqQ_js = robot.nq - 7, robot.nv - 6
        if dimQ_js != rep_Q_js.size:
            raise ConfigException(
                f"{robot_name}'s Pinocchio joint-space dimension {dimQ_js} does not match the joint-space representation"
                f"`Q_js` dimension {rep_Q_js.size}"
            )
        if dimTqQ_js != rep_TqQ_js.size:
            raise ConfigException(
                f"{robot_name}'s Pinocchio joint-space tangent dimension {dimTqQ_js} does not match the joint-space "
                f"tangent representation `TqQ_js` dimension {rep_TqQ_js.size}"
            )

    # Add representations to the group.
    G.representations.update(
        Q_js=rep_Q_js,
        TqQ_js=rep_TqQ_js,
        R3=rep_R3,
        E3=rep_E3,
        R3_pseudo=rep_R3pseudo,
        E3_pseudo=rep_E3pseudo,
        SO3_flat=rep_rot_flat,
    )

    if return_robot:
        return robot, G
    else:
        return G


def generate_euclidean_space_representations(G: Group) -> tuple[Representation, ...]:
    """Generate the E3 representation of the group G.

    This representation is used to transform all members of the Euclidean Space in 3D.
    I.e., points, vectors, pseudo-vectors, etc.
    TODO: List representations generated.

    Args:
        G (Group): Symmetry group of the robot.

    Returns:
        rep_E3 (Representation): Representation of the group G on the Euclidean Space in 3D.
    """
    # Configure E3 representations and group
    if isinstance(G, CyclicGroup):
        if G.order() == 2:  # Reflection symmetry
            rep_R3 = G.irrep(0) + G.irrep(1) + G.trivial_representation
        else:
            rep_R3 = G.irrep(1) + G.trivial_representation
    elif isinstance(G, DihedralGroup):
        rep_R3 = G.irrep(0, 1) + G.irrep(1, 1) + G.trivial_representation
    elif isinstance(G, DirectProductGroup):
        if G.name == "Klein4":
            rep_R3 = G.representations["rectangle"] + G.trivial_representation
            rep_R3 = escnn.group.change_basis(rep_R3, np.array([[0, 1, 0], [1, 0, 0], [0, 0, 1]]), name="R3")
        elif G.name == "FullCylindricalDiscrete":
            rep_hx = np.array(np.array([[-1, 0, 0], [0, 1, 0], [0, 0, 1]]))
            rep_hy = np.array(np.array([[1, 0, 0], [0, -1, 0], [0, 0, 1]]))
            rep_hz = np.array(np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]]))
            rep_E3_gens = {h: rep_h for h, rep_h in zip(G.generators, [rep_hy, rep_hx, rep_hz])}
            rep_E3_gens[G.identity] = np.eye(3)
            rep_R3 = group_rep_from_gens(G, rep_E3_gens)
        else:
            raise NotImplementedError(f"Direct product {G} not implemented yet.")
    else:
        raise NotImplementedError(f"Group {G} not implemented yet.")

    # Representation of unitary/orthogonal transformations in d dimensions.
    rep_R3.name = "R3"

    # We include some utility symmetry representations for different geometric objects.
    # We define a Ed as a (d+1)x(d+1) matrix representing a homogenous transformation matrix in d dimensions.
    rep_E3 = rep_R3 + G.trivial_representation
    rep_E3.name = "E3"

    # Build a representation of orthogonal transformations of pseudo-vectors.
    # That is if det(rep_O3(h)) == -1 [improper rotation] then we have to change the sign of the pseudo-vector.
    # See: https://en.wikipedia.org/wiki/Pseudovector
    pseudo_gens = {h: -1 * rep_R3(h) if np.linalg.det(rep_R3(h)) < 0 else rep_R3(h) for h in G.generators}
    pseudo_gens[G.identity] = np.eye(3)
    rep_R3pseudo = group_rep_from_gens(G, pseudo_gens)
    rep_R3pseudo.name = "R3_pseudo"

    rep_E3pseudo = rep_R3pseudo + G.trivial_representation
    rep_E3pseudo.name = "E3_pseudo"

    return rep_R3, rep_E3, rep_R3pseudo, rep_E3pseudo



import logging
import pathlib
import time
from typing import Callable

import pytorch_lightning as pl
import torch
from escnn.nn import EquivariantModule

from morpho_symm.nn.EMLP import EMLP
from morpho_symm.nn.MLP import MLP
from morpho_symm.utils.mysc import flatten_dict

log = logging.getLogger(__name__)

LossCallable = Callable[
    [
        torch.Tensor,
        torch.Tensor,
    ],
    torch.Tensor,
]
MetricCallable = Callable[
    [
        torch.Tensor,
        torch.Tensor,
    ],
    dict,
]


class LightningModel(pl.LightningModule):
    def __init__(
        self,
        lr,
        loss_fn: LossCallable,
        metrics_fn: MetricCallable,
        test_epoch_metrics_fn=None,
        val_epoch_metrics_fn=None,
        log_preact=False,
        log_w=False,
    ):
        super().__init__()
        # self.model_type = model.__class__.__name__
        self.lr = lr

        # self.model = model
        self._loss_fn = loss_fn
        self.compute_metrics = metrics_fn
        self.test_epoch_metrics_fn = test_epoch_metrics_fn
        self.val_epoch_metrics_fn = val_epoch_metrics_fn
        self._log_w = log_w
        self._log_preact = log_preact
        # Save hyperparams in model checkpoint.
        # TODO: Fix this/home/dordonez/Projects/MorphoSymm/launch/sample_eff
        self.save_hyperparameters()

    def set_model(self, model: [EMLP, MLP]):
        self.model = model
        self.model_type = model.__class__.__name__
        self.equivariant = isinstance(model, EquivariantModule)

    def forward(self, x):
        y = self.model(x)
        return y

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_pred = self.model(x if not self.equivariant else self.model.in_type(x))
        y_pred = y_pred if not self.equivariant else y_pred.tensor
        loss = self._loss_fn(y_pred, y)

        metrics = self.compute_metrics(y_pred, y)
        self.log_metrics(metrics, suffix="train", batch_size=y.shape[0])
        self.log("loss/train", loss, prog_bar=False)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        y_pred = self.model(x if not self.equivariant else self.model.in_type(x))
        y_pred = y_pred if not self.equivariant else y_pred.tensor
        loss = self._loss_fn(y_pred, y)

        metrics = self.compute_metrics(y_pred, y)
        self.log_metrics(metrics, suffix="val", batch_size=y.shape[0])
        self.log("loss/val", loss, prog_bar=False)
        return loss

    def test_step(self, batch, batch_idx):
        x, y = batch
        y_pred = self.model(x if not self.equivariant else self.model.in_type(x))
        y_pred = y_pred if not self.equivariant else y_pred.tensor
        loss = self._loss_fn(y_pred, y)

        metrics = self.compute_metrics(y_pred, y)
        self.log_metrics(metrics, suffix="test", batch_size=y.shape[0])
        self.log("loss/test", loss, prog_bar=False)
        return loss

    def predict_step(self, batch, batch_idx, **kwargs):
        x, y = batch
        return self.model(x)

    def log_metrics(self, metrics: dict, suffix="", batch_size=None):
        flat_metrics = flatten_dict(metrics)
        for k, v in flat_metrics.items():
            name = f"{k}/{suffix}"
            self.log(name, v, prog_bar=False, batch_size=batch_size)

    def on_train_epoch_start(self) -> None:
        self.epoch_start_time = time.time()

    def on_train_epoch_end(self):
        self.log("time_per_epoch", time.time() - self.epoch_start_time, prog_bar=False, on_epoch=True)
        if self._log_w:
            self.log_weights()
        if self._log_preact:
            self.log_preactivations()

    def on_train_end(self) -> None:
        ckpt_call = self.trainer.checkpoint_callback
        if ckpt_call is not None:
            ckpt_path = pathlib.Path(ckpt_call.dirpath).joinpath(
                ckpt_call.CHECKPOINT_NAME_LAST + ckpt_call.FILE_EXTENSION
            )
            best_path = pathlib.Path(ckpt_call.dirpath).joinpath(ckpt_call.filename + ckpt_call.FILE_EXTENSION)
            if ckpt_path.exists() and best_path.exists():
                # Remove last model ckpt leave only best, to hint training successful termination.
                ckpt_path.unlink()
                log.info(f"Removing last ckpt {ckpt_path} from successful training run.")

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.lr)
        return optimizer

    def get_metrics(self):
        # don't show the version number on console logs.
        items = super().get_metrics()
        items.pop("v_num", None)
        return items



"""PinBulletWrapper.

Code based on: https://github.com/machines-in-motion/bullet_utils.

Pybullet interface using pinocchio's convention.

License: BSD 3-Clause License
Copyright (C) 2018-2019, New York University , Max Planck Gesellschaft
Copyright note valid unless otherwise stated in individual files.
All rights reserved.
"""

import logging
from typing import Iterable, Optional

import numpy as np
from pybullet_utils.bullet_client import BulletClient
from robot_descriptions.loaders.pybullet import load_robot_description as pb_load_robot_description

from morpho_symm.robots.PinSimWrapper import JointWrapper, PinSimWrapper, SimPinJointWrapper, State

log = logging.getLogger(__name__)


class PinBulletWrapper(PinSimWrapper):
    """Bridge class between pybullet and pinocchio.

    Attributes:
        nq (int): Dimension of the generalized coordiantes.
        nv (int): Dimension of the generalized velocities.
        n_js (int): Number of joints.
        nf (int): Number of end-effectors.
        robot_id (int): PyBullet id of the robot.
        pinocchio_robot (Pinocchio.RobotWrapper): Pinocchio RobotWrapper for the robot.
        fixed_base (bool): Determines if the robot base if fixed.
    """

    def __init__(
        self,
        robot_name: str,
        endeff_names: Optional[Iterable] = None,
        fixed_base=False,
        reference_robot: Optional["PinBulletWrapper"] = None,
        hip_height=1.0,
        init_q=None,
        q_zero=None,
    ):
        """Initializes the wrapper.

        Args:
            robot_name (str): Name of this robot instance
            endeff_names (List[str]): Names of the end-effectors.
            fixed_base (bool, optional): Determines if the robot base if fixed. Defaults to False.
            reference_robot (PinBulletWrapper, optional): Instance to copy the pinocchio model from. Defaults to None.
            hip_height (float): Height of the hip of the robot. Hip height on resting configuration.
            init_q (List[float]): Initial configuration of the robot of length nq.
            q_zero (List[float]): Zero configuration. This can be different from the Zero config defined in the URDF.
        """
        super().__init__(robot_name, endeff_names, fixed_base, reference_robot, hip_height, init_q, q_zero)

        self._pb = None
        self.world = None
        self.robot_id = None
        self.bullet_endeff_ids = {}
        self.bullet_ids_allowed_floor_contacts = []

    def configure_bullet_simulation(
        self, bullet_client: BulletClient, world=None, base_pos=(0, 0, 0), base_ori=(0, 0, 0, 1)
    ):
        """Configures the bullet simulation and loads this robot URDF description."""
        # Load robot to simulation
        self._pb = bullet_client
        self.world = world
        self.robot_id = self.load_bullet_robot(base_pos, base_ori)
        assert self.robot_id is not None

        log.debug(f"Configuring Bullet Robot for {self.name}")
        bullet_joint_map = {}  # Key: joint name - Value: joint id

        auto_end_eff = False
        if self.endeff_names is None:
            auto_end_eff = True
            self._endeff_names = []

        n_bullet_joints = self.bullet_client.getNumJoints(self.robot_id)
        for bullet_joint_id in range(n_bullet_joints):
            joint_info = self.bullet_client.getJointInfo(self.robot_id, bullet_joint_id)
            joint_name = joint_info[1].decode("UTF-8")
            bullet_joint_map[joint_name] = bullet_joint_id
            if joint_name in self.pin_joint_space:
                pin_joint = self.pin_joint_space[joint_name]
                # Create Joint Wrapper with parameter convention from pybullet
                bullet_joint_type = joint_info[2]
                if bullet_joint_type in [self.bullet_client.JOINT_REVOLUTE, self.bullet_client.JOINT_PRISMATIC]:
                    nq, nv = 1, 1
                elif bullet_joint_type == self.bullet_client.JOINT_SPHERICAL:
                    nq, nv = 4, 3
                elif bullet_joint_type == self.bullet_client.JOINT_FIXED:
                    nq, nv = 0, 0
                elif bullet_joint_type == self.bullet_client.JOINT_PLANAR:
                    nq, nv = 2, 2
                else:
                    raise NotImplementedError(f"Joint type {bullet_joint_type} not handled")
                bullet_joint = JointWrapper(
                    type=bullet_joint_type,
                    idx_q=joint_info[3],
                    idx_v=joint_info[4],
                    nq=nq,
                    nv=nv,
                    pos_limit_low=joint_info[8],
                    pos_limit_high=joint_info[9],
                )
                # Integrate the two convention in a utility class handling the conversions accordingly
                pb_pin_joint = BulletJointWrapper(
                    pin_joint=pin_joint,
                    bullet_joint=bullet_joint,
                    bullet_idx=joint_info[0],
                    bullet_client=self.bullet_client,
                    damping=joint_info[6],
                    friction=joint_info[7],
                    max_force=joint_info[10],
                    max_vel=joint_info[11],
                    axis=joint_info[13],
                    link_name=joint_info[12],
                    parent_frame_pos=joint_info[14],
                    parent_frame_ori=joint_info[15],
                    parent_link_idx=joint_info[16],
                )
                # Store the SimPinJointWrapper for all joint-space joints.
                self.joint_space[joint_name] = pb_pin_joint

            elif auto_end_eff:
                if not np.any([s in joint_name for s in ["base", "imu", "hip", "camera", "accelero", "rotor"]]):
                    log.debug(f"Adding end-effector {joint_name}")
                    self._endeff_names.append(joint_name)
            else:
                log.debug(f"unrecognized joint semantic type {joint_name}")

        # In pybullet, the contact wrench is measured at action joint. In our case
        # the joint is fixed joint. Pinocchio doesn't add fixed joints into the joint
        # list. Therefore, the computation is done wrt to the frame of the fixed joint.
        self.bullet_endeff_ids = {name: bullet_joint_map[name] for name in self.endeff_names}
        self.bullet_ids_allowed_floor_contacts = [bullet_joint_map[name] for name in self.endeff_names]

        # Configure end effectors
        self.pb_nq = 7 + np.sum([self.joint_space[name].sim_joint.nq for name in self.joint_space_names])
        self.pb_nv = 6 + np.sum([self.joint_space[name].sim_joint.nv for name in self.joint_space_names])

        self.nf = len(self.endeff_names)
        self.pinocchio_endeff_ids = {name: self.pinocchio_robot.model.getFrameId(name) for name in self.endeff_names}
        self.bullet_joint_map = bullet_joint_map

    def get_force(self):
        """Returns the force readings as well as the set of active contacts.

        Returns:
            (:obj:`list` of :obj:`int`): List of active contact frame ids.
            (:obj:`list` of np.array((6,1))) List of active contact forces.
        """
        active_contacts_frame_ids = []
        active_bullet_link_id = []

        contact_forces = []
        undesired_contacts = False  # Weather contact with other body parts and floor occured.

        # Get the contact model using the pybullet.getContactPoints() api.
        contact_points = self.bullet_client.getContactPoints(bodyA=self.robot_id, bodyB=self.world.floor_id)

        bullet_endeff_ids = list(self.bullet_endeff_ids.values())
        endeff_names = list(self.bullet_endeff_ids.keys())

        for contact_info in reversed(contact_points):
            robot_link_in_contact = contact_info[3]

            # if robot_link_in_contact not in bullet_endeff_ids:
            if robot_link_in_contact not in self.bullet_ids_allowed_floor_contacts:
                # link_name = self.bullet_client.getJointInfo(self.robot_id, robot_link_in_contact)[12].decode("UTF-8")
                undesired_contacts = True
                # continue
            contact_normal = contact_info[7]
            normal_force = contact_info[9]
            lateral_friction_direction_1 = contact_info[11]
            lateral_friction_force_1 = contact_info[10]
            lateral_friction_direction_2 = contact_info[13]
            lateral_friction_force_2 = contact_info[12]

            if robot_link_in_contact in active_bullet_link_id:
                continue

            if robot_link_in_contact in bullet_endeff_ids:
                active_endeff_name = endeff_names[bullet_endeff_ids.index(robot_link_in_contact)]
                active_contacts_frame_ids.append(self.pinocchio_endeff_ids[active_endeff_name])

            active_bullet_link_id.append(robot_link_in_contact)

            force = np.zeros(6)
            force[:3] = (
                normal_force * np.array(contact_normal)
                + lateral_friction_force_1 * np.array(lateral_friction_direction_1)
                + lateral_friction_force_2 * np.array(lateral_friction_direction_2)
            )

            contact_forces.append(force)

        return active_contacts_frame_ids[::-1], contact_forces[::-1], undesired_contacts

    def get_state_sim(self) -> State:
        """Fetch state from bullet (q, dq).

        Obtains the system state from the simulator [bullet] and returns the generalized positions and velocities in the
        simulator convenction.

        Returns:
            q (ndarray): Generalized position coordinates of shape in bullet convention.
            v (ndarray): Generalized velocity coordinates of shape in bullet convention.
        """
        q_sim = np.zeros(self.pb_nq)
        v_sim = np.zeros(self.pb_nv)

        if not self.fixed_base:
            base_inertia_pos, base_inertia_quat = self.bullet_client.getBasePositionAndOrientation(self.robot_id)
            # Get transform between inertial frame and link frame in base
            base_stat = self.bullet_client.getDynamicsInfo(self.robot_id, -1)
            base_inertia_link_pos, base_inertia_link_quat = self.bullet_client.invertTransform(
                base_stat[3], base_stat[4]
            )
            pos, orn = self.bullet_client.multiplyTransforms(
                base_inertia_pos, base_inertia_quat, base_inertia_link_pos, base_inertia_link_quat
            )

            q_sim[:3] = pos
            q_sim[3:7] = orn

            vel, orn = self.bullet_client.getBaseVelocity(self.robot_id)  # Return in "world" inertial reference frame
            v_sim[:3] = vel
            v_sim[3:6] = orn

            # Pinocchio assumes the base velocity to be in the body frame -> rotate.
            rot_base2world = np.array(self.bullet_client.getMatrixFromQuaternion(q_sim[3:7])).reshape((3, 3))
            v_sim[0:3] = rot_base2world.T.dot(v_sim[0:3])
            v_sim[3:6] = rot_base2world.T.dot(v_sim[3:6])

        # Fetch joint state from bullet
        joint_states = self.bullet_client.getJointStates(
            self.robot_id, [self.joint_space[m].bullet_idx for m in self.joint_space_names]
        )
        for joint_name, joint_state in zip(self.joint_space_names, joint_states):
            joint = self.joint_space[joint_name]
            q_sim[joint.sim_joint.idx_q : joint.sim_joint.idx_q + joint.sim_joint.nq] = joint_state[0]
            v_sim[joint.sim_joint.idx_q : joint.sim_joint.idx_q + joint.sim_joint.nq] = joint_state[1]

        return q_sim, v_sim

    def pin2sim(self, q, v) -> State:
        pb_q = np.zeros(self.pb_nq)
        pb_v = np.zeros(self.pb_nv)

        for joint_name, joint in self.joint_space.items():
            q_joint = q[joint.pin_joint.idx_q : joint.pin_joint.idx_q + joint.pin_joint.nq]
            dq_joint = v[joint.pin_joint.idx_v : joint.pin_joint.idx_v + joint.pin_joint.nv]
            pb_q_joint, pb_v_joint = joint.pin2sim(q_joint, dq_joint)
            # Place value in simulator joint index
            pb_q[joint.sim_joint.idx_q : joint.sim_joint.idx_q + joint.sim_joint.nq] = pb_q_joint
            pb_v[joint.sim_joint.idx_v : joint.sim_joint.idx_v + joint.sim_joint.nv] = pb_v_joint

        # Base configuration
        pb_q[:7] = q[:7]
        pb_v[:6] = v[:6]

        return pb_q, pb_v

    def sim2pin(self, pb_q, pb_v) -> State:
        q = np.zeros(self.nq)
        v = np.zeros(self.nv)
        for joint_name, joint in self.joint_space.items():
            # Extract position and velocity coordinates from simulator
            pb_q_joint = pb_q[joint.sim_joint.idx_q : joint.sim_joint.idx_q + joint.sim_joint.nq]
            pb_v_joint = pb_v[joint.sim_joint.idx_v : joint.sim_joint.idx_v + joint.sim_joint.nv]
            # Convert position and velocity coordinates to pinocchio convention
            q_joint, v_joint = joint.sim2pin(pb_q_joint, pb_v_joint)
            # Place values in Pinocchio joint index ordering
            q[joint.pin_joint.idx_q : joint.pin_joint.idx_q + joint.pin_joint.nq] = q_joint
            v[joint.pin_joint.idx_v : joint.pin_joint.idx_v + joint.pin_joint.nv] = v_joint

        # Base configuration
        q[:7] = pb_q[:7]
        v[:6] = pb_v[:6]

        return q, v

    def reset_state_sim(self, q, v) -> None:
        """Reset robot state in bullet to the described (q, dq) configuration in pinocchio convention.

        Args:
            q (ndarray): Generalized position coordinates of shape in bullet convention.
            v (ndarray): Generalized velocity coordinates of shape in bullet convention.
        """
        assert not np.iscomplexobj(q), "q must be real valued"
        assert not np.iscomplexobj(v), "v must be real valued"

        q, v = np.array(q), np.array(v)

        def vec2list(m):
            return np.asarray(m.T).reshape(-1).tolist()

        for joint_name, joint in self.joint_space.items():
            self.bullet_client.resetJointState(
                self.robot_id,
                joint.bullet_idx,
                q[joint.sim_joint.idx_q : joint.sim_joint.idx_q + joint.sim_joint.nq],
                v[joint.sim_joint.idx_v : joint.sim_joint.idx_v + joint.sim_joint.nv],
            )

        if not self.fixed_base:
            # Get transform between inertial frame and link frame in base
            base_stat = self.bullet_client.getDynamicsInfo(self.robot_id, -1)
            base_pos, base_quat = self.bullet_client.multiplyTransforms(
                vec2list(q[:3]), vec2list(q[3:7]), base_stat[3], base_stat[4]
            )
            self.bullet_client.resetBasePositionAndOrientation(self.robot_id, base_pos, base_quat)

            # Pybullet assumes the base velocity to be aligned with the world frame.
            rot = np.array(self.bullet_client.getMatrixFromQuaternion(q[3:7])).reshape((3, 3))
            self.bullet_client.resetBaseVelocity(self.robot_id, vec2list(rot.dot(v[:3])), vec2list(rot.dot(v[3:6])))

    def load_bullet_robot(self, base_pos=None, base_ori=None) -> int:
        """Function to load and configure the pinocchio instance of your robot.

        Returns:
            int: Bullet robot body id.
        """
        self.robot_id = pb_load_robot_description(
            f"{self.name}_description",
            basePosition=base_pos,
            baseOrientation=base_ori,
            flags=self.bullet_client.URDF_USE_INERTIA_FROM_FILE | self.bullet_client.URDF_USE_SELF_COLLISION,
            useFixedBase=self.fixed_base,
        )
        return self.robot_id

    @property
    def bullet_client(self):
        """Returns the bullet client instance this robot is associated with.

        Returns:
            BulletClient (int): Bullet client instance.
        """
        if self._pb is None:
            raise RuntimeError("Robot bullet simulation not configured, remember to call `configure_bullet_simulation`")
        return self._pb

    def get_base_position_world(self):
        """Returns the position of the base in the world frame.

        Returns:
            np.array((3,)) with the translation
            np.array((4,)) with angular position in quaternion.
        """
        base_inertia_pos, base_inertia_quat = self.bullet_client.getBasePositionAndOrientation(self.robot_id)
        return base_inertia_pos, base_inertia_quat

    def get_base_velocity_world(self):
        """Returns the velocity of the base in the world frame.

        Returns:
            np.array((6,1)) with the translation and angular velocity
        """
        vel, orn = self.bullet_client.getBaseVelocity(self.robot_id)
        return np.array(vel + orn).reshape(6, 1)

    def get_base_acceleration_world(self):
        """Returns the numerically-computed acceleration of the base in the world frame.

        Returns:
            np.array((6,1)) vector of linear and angular acceleration
        """
        return np.concatenate((self.base_linacc, self.base_angacc))

    def __repr__(self):
        """."""
        bullet_id = f"({self.robot_id})" if hasattr(self, "robot_id") else ""
        return f"{self.name}{bullet_id}-nq:{self.nq}-nv:{self.nv}"

    @staticmethod
    def from_instance(other: "PinBulletWrapper") -> "PinBulletWrapper":
        """Creates another instance of this robot wrapper without duplicating the model or data from pinocchio robot.

        This is usefull when we want to spawn multiple instances of the same robot on the physics simulator.


        Args:
            other (PinBulletWrapper): The instance from which to get the model and data for the pinocchio model

        Returns:
            PinBulletWrapper: A new instance of this robot wrapper
        """
        return PinBulletWrapper(
            robot_name=other.name,
            endeff_names=other.endeff_names,
            fixed_base=other.fixed_base,
            # reference_robot=other,
            init_q=other._init_q,
            hip_height=other.hip_height,
        )


class BulletJointWrapper(SimPinJointWrapper):
    """Auxiliary class to integrate Bullet and Pinocchio Joint models."""

    def __init__(
        self,
        pin_joint: JointWrapper,
        bullet_joint: JointWrapper,
        bullet_client: BulletClient,
        bullet_idx: int,
        damping: float,
        friction: float,
        max_force: float,
        max_vel: float,
        link_name: str,
        axis: np.ndarray,
        parent_frame_pos: np.ndarray,
        parent_frame_ori: np.ndarray,
        parent_link_idx: int,
    ):
        self.pin_joint = pin_joint
        self.sim_joint = bullet_joint
        self.bullet_client = bullet_client
        self.bullet_idx = bullet_idx
        self.damping = damping
        self.friction = friction
        self.max_force = max_force
        self.max_vel = max_vel
        self.link_name = link_name
        self.axis = axis
        self.parent_frame_pos = parent_frame_pos
        self.parent_frame_ori = parent_frame_ori
        self.parent_link_idx = parent_link_idx

    def sim2pin(self, q, v):
        if self.pin_joint.nq == 1:
            return q, v
        if self.pin_joint.nq == 2 and self.pin_joint.nv == 1:  # Unit circle
            return np.asarray([np.cos(q), np.sin(q)]).flatten(), v
        else:
            raise NotImplementedError()

    def pin2sim(self, q, v):
        if self.pin_joint.nq == 1:
            return q, v
        if self.pin_joint.nq == 2 and self.pin_joint.nv == 1:  # Unit circle
            theta = np.arctan2(q[1], q[0])
            return np.asarray([theta]).flatten(), v
        else:
            raise NotImplementedError()
